import unittest
from typing import Optional
from unittest import mock

from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Serializable
from jivago.lang.registration import Registration
from jivago.lang.registry import Registry
from jivago.lang.stream import Stream
from jivago.serialization.deserializer import Deserializer
from jivago.serialization.serialization_exception import SerializationException
from jivago.wsgi.annotations import Resource, Path
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.invocation.incorrect_resource_parameters_exception import IncorrectResourceParametersException
from jivago.wsgi.invocation.parameters import PathParam, QueryParam, OptionalQueryParam
from jivago.wsgi.invocation.route_handler_factory import RouteHandlerFactory
from jivago.wsgi.methods import GET, POST
from jivago.wsgi.request.headers import Headers
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response
from jivago.wsgi.routing.cors.cors_request_handler_factory import CorsRequestHandlerFactory
from jivago.wsgi.routing.route_registration import RouteRegistration
from jivago.wsgi.routing.routing_rule import RoutingRule
from jivago.wsgi.routing.routing_table import RoutingTable
from jivago.wsgi.routing.table.reflective_routing_table import ReflectiveRoutingTable
from test_utils.request_builder import RequestBuilder

BODY = {"key": "value"}
DTO_BODY = {"name": "hello"}

PATH = 'path'
A_PATH_PARAM = "a-param"

HTTP_METHOD = GET


# TODO Remove this integrated test class when rewriting resource invoker

class TestIntegratedResourceInvoker(object):
    def __init__(self, service_locator: ServiceLocator, routing_table: RoutingTable,
                 deserializer: Deserializer):
        self.deserializer = deserializer
        self.service_locator = service_locator
        self.routing_table = routing_table

    def invoke(self, request: Request) -> Response:
        response = ResourceClass.the_response
        FilterChain([], RouteHandlerFactory(self.service_locator, self.deserializer,
                                            [RoutingRule("", self.routing_table)],
                                            CorsRequestHandlerFactory([]))).doFilter(request, response)

        return response


class IntegratedResourceInvokerTest(unittest.TestCase):

    def setUp(self):
        self.serviceLocator = ServiceLocator()
        self.serviceLocator.bind(ResourceClass, ResourceClass)
        registry = Registry()
        self.routingTable = ReflectiveRoutingTable(registry,
                                                   [Registration(ResourceClass, arguments={"value": PATH})])
        self.deserializer = Deserializer(registry)
        self.resource_invocator = TestIntegratedResourceInvoker(self.serviceLocator, self.routingTable,
                                                                self.deserializer)
        self.request = Request('GET', PATH, {}, "", "")
        ResourceClass.has_been_called = False

    def test_whenInvoking_thenGetsCorrespondingResourceInRoutingTable(self):
        self.resource_invocator.invoke(self.request)

        self.assertTrue(ResourceClass.has_been_called)

    def test_givenRouteWhichRequiresARequestObject_whenInvoking_thenInvokeTheRouteWithTheRequestObject(self):
        self.request.path += "/request"
        self.resource_invocator.invoke(self.request)

        # Assert is done in the route method below

    def test_givenRouteWhichReturnsResponse_whenInvoking_thenDirectlyReturnTheRouteResponse(self):
        response = self.resource_invocator.invoke(self.request)

        self.assertIsInstance(response, Response)

    def test_givenRouteWhichReturnsSomethingOtherThanAResponse_whenInvoking_thenReturnValueIsSavedInTheBodyOfTheResponse(
            self):
        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(ResourceClass().a_method(), response.body)

    def test_givenRouteWhichReturnsSomethingOtherThanAResponse_whenInvoking_thenResponseHas200OkStatus(self):
        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(200, response.status)

    def test_givenRouteWhichTakesADictionaryAsParameter_whenInvoking_thenPassRequestBodyAsParameter(self):
        self.request = Request('POST', PATH + "/dictionary", {}, "", BODY)

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(ResourceClass.the_response, response)

    def test_givenRouteWhichTakesADtoAsParameter_whenInvoking_thenDeserializeRequestBodyIntoTheDtoClass(self):
        self.request = Request('POST', PATH + "/dto", {}, "", DTO_BODY)

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(ResourceClass.the_response, response)

    def test_givenRouteWhichTakesADtoWithoutAnExplicitConstructor_whenInvoking_thenDeserializeRequestBodyIntoDto(self):
        self.request = Request('POST', PATH + "/dto-no-param", {}, "", DTO_BODY)

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(ResourceClass.the_response, response)

    def test_givenRouteWithPathParameters_whenInvoking_thenPassAsArguments(self):
        self.request = Request('GET', PATH + "/path-param/" + A_PATH_PARAM, {}, "", "")

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(A_PATH_PARAM, response.body)

    def test_givenRouteWithNumericPathParameter_whenInvoking_thenParseStringToNumberBeforePassing(self):
        a_numeric_path_param = 5
        self.request = Request('GET', PATH + "/numeric-param/" + str(a_numeric_path_param), {}, "", "")

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(a_numeric_path_param, response.body)

    def test_givenRouteWithQueryParameter_whenInvoking_thenParseQueryParametersBeforeInvoking(self):
        name = "foobar"
        self.request = Request('GET', PATH + "/query-param", {}, "name=" + name, "")

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(name, response.body)

    def test_givenEscapedQueryParameter_whenInvoking_thenUnescapeBeforeInvoking(self):
        escaped_name = "rocket%20man"
        unescaped_name = "rocket man"
        self.request = Request('GET', PATH + "/query-param", {}, "name=" + escaped_name, "")

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(unescaped_name, response.body)

    def test_givenEscapedPathParameter_whenInvoking_thenUnescapeBeforeInvoking(self):
        escaped_name = "rocket%20man"
        unescaped_name = "rocket man"
        self.request = Request('GET', PATH + "/path-param/" + escaped_name, {}, "", "")

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(unescaped_name, response.body)

    def test_givenIncorrectParameters_whenInvoking_thenThrowIncorrectResourceParametersException(self):
        self.request = Request('POST', PATH + "/dto", {}, "", "")

        with self.assertRaises(IncorrectResourceParametersException):
            self.resource_invocator.invoke(self.request)

    def test_givenOverloadedRouteRegistrations_whenInvoking_thenChooseMethodBasedOnMethodSignature(self):
        self.request = Request('GET', PATH + "/overloaded", {}, "query=foo", "")

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(ResourceClass.OVERLOADED_RETURN, response.body)

    def test_givenMissingOptionalResourceArgument_whenInvoking_thenCallEvenIfTheArgumentIsMissing(self):
        self.request = RequestBuilder().path(PATH + "/nullable-query").build()

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual(None, response.body)

    def test_givenPresentOptionalResourceArgument_whenInvoking_thenCallResourceWithParameter(self):
        self.request = RequestBuilder().path(PATH + "/nullable-query").query_string("query=foo").build()

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual("foo", response.body)

    def test_givenApplicationException_whenInvokingResource_thenLetTheExceptionRise(self):
        self.request = RequestBuilder().path(PATH + "/error").build()

        with self.assertRaises(AnException):
            self.resource_invocator.invoke(self.request)

    def test_givenResourceWithHeadersParameter_whenInvokingResource_thenInvokeRoutingMethodWithHeadersObject(self):
        self.request = RequestBuilder().path(PATH + "/headers").headers({"foo": "bar"}).build()

        response = self.resource_invocator.invoke(self.request)

        self.assertEqual([("foo", "bar")], Stream(response.body).toList())

    def test_givenSerializationError_whenInvokingResource_thenRaiseException(self):
        self.resource_invocator.deserializer = mock.create_autospec(Deserializer)
        self.resource_invocator.deserializer.deserialize.side_effect = SerializationException()
        self.request = RequestBuilder().method("POST").path(PATH + "/dto").build()

        with self.assertRaises(IncorrectResourceParametersException):
            self.resource_invocator.invoke(self.request)


@Serializable
class A_Dto(object):
    name: str

    def __init__(self, name: str):
        self.name = name


@Serializable
class ADtoWithoutAnExplicitConstructor(object):
    name: str

    def a_dummy_function(self) -> bool:
        return False


class AnException(Exception):
    pass


@Resource(PATH)
class ResourceClass(object):
    has_been_called = False

    the_response = Response(402, {}, "a response")

    @GET
    def a_method(self):
        ResourceClass.has_been_called = True
        return "hello"

    @GET
    @Path("/request")
    def a_method_which_requires_a_request(self, request: Request) -> Response:
        assert isinstance(request, Request)
        return self.the_response

    @POST
    @Path("/dictionary")
    def a_method_which_requires_a_dictionary(self, body: dict) -> Response:
        assert isinstance(body, dict)
        assert body == BODY
        return self.the_response

    @POST
    @Path("/dto")
    def a_method_which_requires_a_dto(self, body: A_Dto) -> Response:
        assert isinstance(body, A_Dto)
        assert body.name == DTO_BODY['name']
        return self.the_response

    @POST
    @Path("/dto-no-param")
    def a_method_which_requires_a_dto_without_a_constructor(self, body: ADtoWithoutAnExplicitConstructor) -> Response:
        assert isinstance(body, ADtoWithoutAnExplicitConstructor)
        assert body.name == DTO_BODY['name']
        assert body.a_dummy_function() == False  # assert function does not get overwritten
        return self.the_response

    @GET
    @Path("/return-dto")
    def returns_a_dto(self) -> A_Dto:
        return A_Dto("a_name")

    @GET
    @Path("/path-param/{name}")
    def with_path_param(self, name: PathParam[str]) -> str:
        assert isinstance(name, str)
        return name

    @GET
    @Path("/numeric-param/{number}")
    def with_numeric_path_param(self, number: PathParam[int]) -> int:
        assert isinstance(number, int)
        return number

    @GET
    @Path("/query-param")
    def with_query_param(self, name: QueryParam[str]) -> str:
        assert isinstance(name, str)
        return name

    @GET
    @Path("/overloaded")
    def overloaded_param(self, name: QueryParam[str]) -> int:
        return 5

    OVERLOADED_RETURN = 6

    @GET
    @Path("/overloaded")
    def overloaded_without_name_parameter(self, query: QueryParam[str]) -> int:
        return self.OVERLOADED_RETURN

    @GET
    @Path("/nullable-query")
    def nullable_query(self, query: OptionalQueryParam[str]) -> Optional[str]:
        return query

    @GET
    @Path("/error")
    def raises_error(self) -> str:
        raise AnException

    @GET
    @Path("/headers")
    def get_with_headers(self, headers: Headers) -> list:
        return headers.items()


ROUTE_REGISTRATION = RouteRegistration(ResourceClass, ResourceClass.a_method, [""], GET)



from jivago.lang.registry import Registry
from jivago.scheduling.annotations import Scheduled
from jivago.scheduling.cron_schedule import CronSchedule
from jivago.scheduling.regular_interval_schedule import RegularIntervalSchedule
from jivago.scheduling.task_scheduler import TaskScheduler


class TaskScheduleInitializer(object):

    def __init__(self, registry: Registry, root_package_name: str):
        self.registry = registry
        self.root_package_name = root_package_name

    def initialize_task_scheduler(self, task_scheduler: TaskScheduler):

        for registration in self.registry.get_annotated_in_package(Scheduled, self.root_package_name):
            schedule = None
            start_time = registration.arguments.get('start')

            if registration.arguments.get('cron'):
                schedule = CronSchedule(registration.arguments['cron'], start_time)

            if registration.arguments.get('every'):
                schedule = RegularIntervalSchedule(registration.arguments['every'], start_time)

            task_scheduler.schedule_task(registration.registered, schedule)



from jivago.lang.registry import Registry
from jivago.wsgi.annotations import Resource
from jivago.wsgi.routing.table.reflective_routing_table import ReflectiveRoutingTable


class AutoDiscoveringRoutingTable(ReflectiveRoutingTable):
    """Reads @Resource annotations to build the routing table."""

    def __init__(self, registry: Registry, root_package_name: str):
        resources = registry.get_annotated_in_package(Resource, root_package_name)
        super().__init__(registry, resources)




import unittest
from typing import Callable

from jivago.wsgi.methods import GET
from jivago.wsgi.routing.route_registration import RouteRegistration

HTTP_METHOD = GET

A_RESOURCE_CLASS: type = None
A_ROUTE_FUNCTION: Callable = None
A_PATH_WITH_PARAMETERS = ["hello", "{name}"]
A_PATH_WITHOUT_PARAMETERS = ["hello", "goodbye"]


class RouteRegistrationTest(unittest.TestCase):

    def setUp(self):
        self.routeRegistration = RouteRegistration(A_RESOURCE_CLASS, A_ROUTE_FUNCTION, A_PATH_WITH_PARAMETERS,
                                                   HTTP_METHOD)

    def test_givenAPathWithParameters_whenParsingPathParameters_thenReturnAllMatchingTokensAsParameters(self):
        gotten_path = "/hello/paul-atreides"

        parameters = self.routeRegistration.parse_path_parameters(gotten_path)

        self.assertEqual(1, len(parameters))
        self.assertEqual("paul-atreides", parameters['name'])

    def test_givenAPathWithoutParameters_whenParsingPathParameters_thenReturnEmptyDictionary(self):
        self.routeRegistration = RouteRegistration(A_RESOURCE_CLASS, A_ROUTE_FUNCTION, A_PATH_WITHOUT_PARAMETERS,
                                                   HTTP_METHOD)
        gotten_path = "/hello/goodbye"

        parameters = self.routeRegistration.parse_path_parameters(gotten_path)

        self.assertEqual(0, len(parameters))



import unittest

from jivago.wsgi.invocation.rewrite.modified_path_flyweight_request import ModifiedPathFlyweightRequest
from test_utils.request_builder import RequestBuilder

NEW_PATH = "/new-path"


class ModifiedPathFlyweightRequestTest(unittest.TestCase):

    def setUp(self):
        self.base_request = RequestBuilder().path("/api/path").build()
        self.modified_request = ModifiedPathFlyweightRequest(self.base_request, NEW_PATH)

    def test_whenGettingPath_thenReturnOverriddenPath(self):
        self.assertEqual(NEW_PATH, self.modified_request.path)

    def test_whenGettingOtherAttribute_thenGetFromBaseRequest(self):
        self.assertEqual(self.base_request.method, self.modified_request.method)
        self.assertEqual(self.base_request.headers, self.modified_request.headers)



import os
from typing import Tuple

from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class PartialContentHandler(object):

    def handle_partial_content_request(self, request: Request, filepath: str, *, max_block_size: int = 2000000) -> Response:
        if request.headers['Range']:
            total_filesize = os.path.getsize(filepath)
            start, end = self._parse_range_string(request.headers['Range'], max_block_size, total_filesize)
            return Response(206, {"Content-Type": "application/octet-stream", "Accept-Ranges": "bytes",
                                  "Content-Range": f"bytes {start}-{end - 1}/{total_filesize}"}, self._read_partial_file(filepath, start, end))

        return Response(200, {"Content-Type": "application/octet-stream"}, self._read_whole_file(filepath))

    def _parse_range_string(self, range_string: str, default_block_size: int, file_end_offset: int) -> Tuple[int, int]:
        start = int(range_string.split("=")[1].split("-")[0])
        end_string = range_string.split("-")[1]
        end = int(end_string) if end_string != '' else start + default_block_size

        if end > file_end_offset:
            end = file_end_offset

        return start, end

    def _read_partial_file(self, filepath: str, start: int, end: int) -> bytes:
        with open(filepath, 'rb') as f:
            f.seek(start)
            return f.read(end - start)

    def _read_whole_file(self, filepath: str) -> bytes:
        with open(filepath, 'rb') as f:
            return f.read()



from jivago.lang.annotations import Override
from jivago.wsgi.invocation.rewrite.modified_path_flyweight_request import ModifiedPathFlyweightRequest
from jivago.wsgi.invocation.route_handler import RouteHandler
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class PathRewritingRouteHandlerDecorator(RouteHandler):

    def __init__(self, decorated: RouteHandler, new_path: str):
        self.decorated = decorated
        self.new_path = new_path

    @Override
    def invoke(self, request: Request) -> Response:
        return self.decorated.invoke(ModifiedPathFlyweightRequest(request, self.new_path))



from typing import Callable

from jivago.event.dispatch.message_dispatcher_function import MessageDispatcherFunction
from jivago.lang.annotations import Override


class MessageDispatcherClass(MessageDispatcherFunction):

    def __init__(self, message_name: str, method: Callable):
        super().__init__(message_name, method)

    @Override
    def _requires_payload_parameter(self) -> bool:
        return self.function.__code__.co_argcount == 2



import unittest

from jivago.wsgi.routing.cors.cors_request_handler_factory import CorsRequestHandlerFactory
from jivago.config.router.cors_rule import CorsRule
from jivago.wsgi.routing.cors.no_matching_cors_rule_exception import NoMatchingCorsRuleException


class CorsRequestHandlerFactoryTest(unittest.TestCase):

    def setUp(self):
        self.rules = [
            CorsRule("/", {"value": "default_rule"}),
            CorsRule("/foo/bar", {"value": "specialized_rule"})
        ]

        self.cors_request_handler_factory = CorsRequestHandlerFactory(self.rules)

    def test_givenMultipleMatchingRules_whenCreatingCorsRequestHandler_thenCreateBasedOnLongestRule(self):
        handler = self.cors_request_handler_factory.create_cors_preflight_handler("/foo/bar")

        self.assertEqual("specialized_rule", handler.cors_headers['value'])

    def test_givenNoMatchingRule_whenCreatingCorsRequestHandler_thenRaiseException(self):
        self.cors_request_handler_factory = CorsRequestHandlerFactory([])

        with self.assertRaises(NoMatchingCorsRuleException):
            self.cors_request_handler_factory.create_cors_preflight_handler("/baz")



import re

from jivago.lang.stream import Stream
from jivago.wsgi.request.headers import Headers
from jivago.wsgi.request.request import Request


class RequestFactory(object):

    def build_request(self, env: dict) -> Request:
        """Builds the Jivago request object from the WSGI environment dictionary."""
        raw_headers = Stream(env.items()).filter(lambda key, value: key.startswith("HTTP_")).map(
            lambda key, value: (re.sub("^HTTP_", "", key).title().replace("_", "-"), value)).toDict()
        request_size = int(env.get('CONTENT_LENGTH')) if 'CONTENT_LENGTH' in env else 0

        if 'CONTENT_TYPE' in env:
            raw_headers['Content-Type'] = env['CONTENT_TYPE']

        return Request(env['REQUEST_METHOD'], env['PATH_INFO'], Headers(raw_headers), env['QUERY_STRING'],
                       env['wsgi.input'].read(request_size))



from typing import Callable

from jivago.event.dispatch.message_dispatcher_class import MessageDispatcherClass
from jivago.inject.service_locator import ServiceLocator


class JitMessageDispatcherClass(MessageDispatcherClass):

    def __init__(self, message_name: str, clazz: type, method: Callable, service_locator: ServiceLocator):
        super().__init__(message_name, method)
        self.method = method
        self.clazz = clazz
        self.service_locator = service_locator

    def handle(self, payload: object):
        instance = self.service_locator.get(self.clazz)

        if self._requires_payload_parameter():
            return self.method(instance, payload)
        return self.method(instance)






from typing import List

from jivago.lang.registration import Registration
from jivago.lang.registry import Registry
from jivago.lang.stream import Stream
from jivago.wsgi.annotations import Path
from jivago.wsgi.methods import http_methods
from jivago.wsgi.routing.table.tree_routing_table import TreeRoutingTable


class ReflectiveRoutingTable(TreeRoutingTable):

    def __init__(self, registry: Registry, resources: List[Registration]):
        super().__init__()
        for resource in resources:
            for primitive in http_methods:
                routable_functions = registry.get_annotated_in_package(primitive, resource.registered.__module__)
                sub_paths = registry.get_annotated_in_package(Path, resource.registered.__module__)

                for route_function in routable_functions:
                    route_sub_path = Stream(sub_paths).firstMatch(lambda r: r.registered == route_function.registered)
                    resource_path = resource.arguments['value']

                    if not route_sub_path.isPresent():
                        self.register_route(primitive, resource_path, resource.registered,
                                            route_function.registered)
                    else:
                        sub_path = route_sub_path.get().arguments['value']
                        path = resource_path + sub_path if \
                            resource_path.endswith('/') or sub_path.startswith('/') \
                            else resource_path + '/' + sub_path

                        self.register_route(primitive, path, resource.registered, route_function.registered)



from typing import List

from jivago.config.router.filtering.filtering_rule import FilteringRule
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.stream import Stream
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.invocation.route_handler_factory import RouteHandlerFactory
from jivago.wsgi.request.request import Request


class FilterChainFactory(object):

    def __init__(self, filtering_rules: List[FilteringRule],
                 service_locator: ServiceLocator,
                 route_handler_factory: RouteHandlerFactory):
        self.route_handler_factory = route_handler_factory
        self.service_locator = service_locator
        self.filtering_rules = filtering_rules

    def create_filter_chain(self, request: Request) -> FilterChain:
        if request.method == 'OPTIONS':
            filters = []
        else:
            filters = Stream(self.filtering_rules) \
                .filter(lambda rule: rule.matches(request.path)) \
                .map(lambda rule: rule.get_filters(self.service_locator)) \
                .flat() \
                .toList()

        return FilterChain(filters, self.route_handler_factory)



from typing import List, Optional

from jivago.config.router.cors_rule import CorsRule
from jivago.lang.stream import Stream
from jivago.wsgi.invocation.route_handler import RouteHandler
from jivago.wsgi.routing.cors.cors_preflight_request_handler import CorsPreflightRequestHandler
from jivago.wsgi.routing.cors.cors_route_handler_decorator import CorsRouteHandlerDecorator
from jivago.wsgi.routing.cors.no_matching_cors_rule_exception import NoMatchingCorsRuleException


class CorsRequestHandlerFactory(object):

    def __init__(self, cors_rules: List[CorsRule]):
        self.cors_rules = cors_rules

    def create_cors_preflight_handler(self, path: str) -> CorsPreflightRequestHandler:
        highest_priority_rule = self._find_highest_priority_rule(path)

        if highest_priority_rule:
            return CorsPreflightRequestHandler(highest_priority_rule.cors_headers)

        raise NoMatchingCorsRuleException()

    def _find_highest_priority_rule(self, path: str) -> Optional[CorsRule]:
        return Stream(self.cors_rules) \
            .filter(lambda rule: rule.matches(path)) \
            .reduce(None, lambda highest, e: e if highest is None or e.takes_precedence_over(highest) else highest)

    def apply_cors_rules(self, path: str, route_handler: RouteHandler) -> RouteHandler:
        rule = self._find_highest_priority_rule(path)
        if rule:
            return CorsRouteHandlerDecorator(rule, route_handler)

        return route_handler



import unittest
from unittest import mock

from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.http_form_deserialization_filter import HttpFormDeserializationFilter
from test_utils.request_builder import RequestBuilder

FORM_BODY = b"key=value"

PARSED_FORM_BODY = {"key": "value"}


class HttpFormDeserializationFilterTest(unittest.TestCase):

    def setUp(self):
        self.filterChainMock: FilterChain = mock.create_autospec(FilterChain)

        self.formDeserializationFilter = HttpFormDeserializationFilter()

    def test_givenRequestWithContentTypWwwForm_whenApplyingFilter_thenBodyGetsParsed(self):
        request = RequestBuilder().method('POST').headers({'CONTENT-TYPE': 'application/x-www-form-urlencoded'}).body(FORM_BODY).build()

        self.formDeserializationFilter.doFilter(request, None, self.filterChainMock)

        self.assertEqual(PARSED_FORM_BODY, request.body)

    def test_givenRequestWithoutAForm_whenApplyingFilter_thenDoNothing(self):
        request = RequestBuilder().method("POST").body(FORM_BODY).build()

        self.formDeserializationFilter.doFilter(request, None, self.filterChainMock)

        self.assertEqual(FORM_BODY, request.body)



import unittest

from jivago.event.config.annotations import EventHandler
from jivago.event.config.runnable_event_handler_binder import RunnableEventHandlerBinder
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.registry import Registry
from jivago.lang.runnable import Runnable


class RunnableEventHandlerBinderTest(unittest.TestCase):

    def setUp(self):
        self.registry = Registry()
        self.service_locator = ServiceLocator()

    def test_whenConfiguringServiceLocator_thenFindAnnotatedEventHandlerRunnableClasses(self):
        RunnableEventHandlerBinder("", self.registry).bind(self.service_locator)

        instance = self.service_locator.get(_RunnableHandler)

        self.assertIsInstance(instance, _RunnableHandler)


@EventHandler("event")
class _RunnableHandler(Runnable):
    pass



from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class RouteHandler(object):

    def invoke(self, request: Request) -> Response:
        raise NotImplementedError



from typing import List, Iterable

from jivago.config.router.router_config_rule import RouterConfigRule
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.stream import Stream
from jivago.serialization.deserializer import Deserializer
from jivago.wsgi.invocation.parameter_selection.parameter_selector_chain import ParameterSelectorChain
from jivago.wsgi.invocation.resource_invoker import ResourceInvoker
from jivago.wsgi.invocation.rewrite.path_rewriting_route_handler_decorator import PathRewritingRouteHandlerDecorator
from jivago.wsgi.invocation.route_handler import RouteHandler
from jivago.wsgi.request.request import Request
from jivago.wsgi.routing.route_registration import RouteRegistration
from jivago.wsgi.routing.routing_table import RoutingTable


class RoutingRule(RouterConfigRule):

    def __init__(self, prefix_path: str, routing_table: RoutingTable, rewrite_path: bool = True):
        self.should_rewrite_path = rewrite_path
        self.prefix_path = prefix_path.rstrip("/")
        self.routing_table = routing_table

    def matches(self, path: str) -> bool:
        return path.startswith(self.prefix_path)

    def get_route_registrations(self, path: str) -> List[RouteRegistration]:
        if self.matches(path):
            return self.routing_table.get_route_registrations(self._truncate_path(path))
        else:
            return []

    def _truncate_path(self, path: str) -> str:
        return path[len(self.prefix_path):]

    def create_route_handlers(self, request: Request,
                              service_locator: ServiceLocator,
                              deserializer: Deserializer) -> Iterable[RouteHandler]:
        
        resource_invokers = Stream(self.get_route_registrations(request.path)) \
            .filter(lambda route: route.http_method == request.method_annotation) \
            .map(lambda route: ResourceInvoker(route,
                                               service_locator,
                                               ParameterSelectorChain(route, deserializer)))
        if self.should_rewrite_path:
            return resource_invokers \
                .map(lambda invoker: PathRewritingRouteHandlerDecorator(invoker, self._truncate_path(request.path)))
        else:
            return resource_invokers



import unittest

from jivago.config.router.cors_rule import CorsRule


class CorsRuleTest(unittest.TestCase):

    def setUp(self):
        self.rule = CorsRule("/", {})

    def test_givenRootPathRule_whenMatching_thenRuleMatchesAnyRequest(self):
        self.assertTrue(self.rule.matches("/"))
        self.assertTrue(self.rule.matches("/foo/bar"))

    def test_whenComparingCorsRules_thenLongerPathRulesTakePrecedenceOverShorterOnes(self):
        self.longer_rule = CorsRule("/foobar", {})

        takes_precedence = self.longer_rule.takes_precedence_over(self.rule)

        self.assertTrue(takes_precedence)

    def test_givenSameLengthRule_whenComparingCorsRules_thenPriorityIsGivenToMostRecentRule(self):
        takes_precedence = self.rule.takes_precedence_over(self.rule)

        self.assertTrue(takes_precedence)



from typing import List, Callable

from jivago.event.config.annotations import EventHandler, EventHandlerClass
from jivago.event.dispatch.jit_message_dispatcher_class import JitMessageDispatcherClass
from jivago.event.dispatch.message_dispatcher import MessageDispatcher
from jivago.event.dispatch.message_dispatcher_function import MessageDispatcherFunction
from jivago.event.dispatch.message_dispatcher_runnable import MessageDispatcherRunnable
from jivago.event.event_bus import EventBus
from jivago.event.synchronous_event_bus import SynchronousEventBus
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.registry import Registry
from jivago.lang.runnable import Runnable
from jivago.lang.stream import Stream

EVENT_NAME_PARAMETER = 'event_name'


class ReflectiveEventBusInitializer(object):

    def __init__(self, service_locator: ServiceLocator, registry: Registry, root_package_name: str = ""):
        self.root_package_name = root_package_name
        self.service_locator = service_locator
        self.registry = registry

    def create_message_bus(self) -> EventBus:
        return SynchronousEventBus(self._find_dispatchers(self.registry, self.root_package_name, self.service_locator))

    def _find_dispatchers(self, registry: Registry,
                          root_package_name: str,
                          service_locator: ServiceLocator) -> List[MessageDispatcher]:
        dispatchers = []
        registrations = registry.get_annotated_in_package(EventHandler, root_package_name)

        for registration in registrations:
            if registration.is_function_registration():
                dispatchers.append(MessageDispatcherFunction(registration.arguments[EVENT_NAME_PARAMETER],
                                                             registration.registered))

            elif registration.is_class_registration() and issubclass(registration.registered, Runnable):
                dispatchers.append(MessageDispatcherRunnable(registration.arguments[EVENT_NAME_PARAMETER],
                                                             registration.registered,
                                                             self.service_locator))

        for registration in registry.get_annotated_in_package(EventHandlerClass, root_package_name):
            registered_class = registration.registered

            for handler_function in registry.get_annotated_in_package(EventHandler, registered_class.__module__):
                if self._is_function_defined_inside_class(handler_function.registered, registered_class):
                    dispatchers.append(JitMessageDispatcherClass(handler_function.arguments[EVENT_NAME_PARAMETER],
                                                                 registered_class,
                                                                 handler_function.registered,
                                                                 service_locator))

        return dispatchers

    def _is_function_defined_inside_class(self, function: Callable, clazz: type) -> bool:
        return function.__qualname__.startswith(clazz.__qualname__)



from jivago.lang.annotations import Override
from jivago.wsgi.filter.system_filters.error_handling.exception_mapper import ExceptionMapper
from jivago.wsgi.request.response import Response
from jivago.wsgi.routing.exception.unknown_path_exception import UnknownPathException


class UnknownPathExceptionMapper(ExceptionMapper):

    @Override
    def handles(self, exception: Exception) -> bool:
        return isinstance(exception, UnknownPathException)

    @Override
    def create_response(self, exception: Exception) -> Response:
        return Response(404, {}, {"message": "Resource Not Found"})



from jivago.wsgi.request.request import Request


class ParameterSelector(object):

    def matches(self, paramater_declaration: type) -> bool:
        raise NotImplementedError

    def format_parameter(self, parameter_name: str, parameter_declaration: type, request: Request) -> object:
        raise NotImplementedError



import re
from typing import List, Union, Type

from jivago.config.router.router_config_rule import RouterConfigRule
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.stream import Stream
from jivago.wsgi.filter.filter import Filter


class FilteringRule(RouterConfigRule):

    def __init__(self, url_pattern: str, filters: List[Union[Filter, Type[Filter]]], regex_pattern: str = None):
        self.filters = filters
        self.regex_matcher = re.compile("^" + url_pattern.replace(".", "\.").replace("*", ".*") + "$") \
            if regex_pattern is None else re.compile(regex_pattern)

    def matches(self, path: str) -> bool:
        return self.regex_matcher.search(path) is not None

    def get_filters(self, service_locator: ServiceLocator) -> List[Filter]:
        return Stream(self.filters).map(lambda f: f if isinstance(f, Filter) else service_locator.get(f)).toList()



from typing import TypeVar, Type

from jivago.lang.registry import Registry
from jivago.lang.stream import Stream
from jivago.serialization.deserialization.built_in_type_deserialization_strategy import \
    BuiltInTypeDeserializationStrategy
from jivago.serialization.deserialization.datetime_deserialization_strategy import DatetimeDeserializationStrategy
from jivago.serialization.deserialization.dictionary_deserialization_strategy import DictionaryDeserializationStrategy
from jivago.serialization.deserialization.list_deserialization_strategy import ListDeserializationStrategy
from jivago.serialization.deserialization.named_tuple_deserialization_strategy import NamedTupleDeserializationStrategy
from jivago.serialization.deserialization.optional_attribute_deserialization_strategy import \
    OptionalAttributeDeserializationStrategy
from jivago.serialization.deserialization.registered_serializable_type_deserialization_strategy import \
    RegisteredSerializableTypeDeserializationStrategy
from jivago.serialization.deserialization.tuple_deserialization_strategy import TupleDeserializationStrategy
from jivago.serialization.deserialization.typed_dictionary_deserialization_strategy import \
    TypedDictionaryDeserializationStrategy
from jivago.serialization.deserialization.typed_list_deserialization_strategy import TypedListDeserializationStrategy
from jivago.serialization.deserialization.typed_tuple_deserialization_strategy import TypedTupleDeserializationStrategy
from jivago.serialization.serialization_exception import SerializationException
from jivago.wsgi.invocation.incorrect_attribute_type_exception import IncorrectAttributeTypeException

T = TypeVar('T')


class Deserializer(object):

    def __init__(self, registry: Registry):
        self.deserialization_strategies = [
            BuiltInTypeDeserializationStrategy(),
            DatetimeDeserializationStrategy(),
            NamedTupleDeserializationStrategy(self),
            TypedDictionaryDeserializationStrategy(self),
            OptionalAttributeDeserializationStrategy(self),
            TypedListDeserializationStrategy(self),
            TypedTupleDeserializationStrategy(self),
            RegisteredSerializableTypeDeserializationStrategy(registry, self),
            DictionaryDeserializationStrategy(),
            ListDeserializationStrategy(),
            TupleDeserializationStrategy()
        ]

    def is_deserializable_type(self, object_clazz: type) -> bool:
        return Stream(self.deserialization_strategies) \
            .anyMatch(lambda s: s.can_handle_deserialization(object_clazz))

    def deserialize(self, obj: dict, object_clazz: Type[T]) -> T:
        try:
            return Stream(self.deserialization_strategies). \
                firstMatch(lambda s: s.can_handle_deserialization(object_clazz)) \
                .orElseThrow(SerializationException) \
                .deserialize(obj, object_clazz)
        except (AttributeError, TypeError):
            raise IncorrectAttributeTypeException()



import os
import unittest

import test_data.test_static
from jivago.wsgi.routing.serving.static_file_routing_table import StaticFileRoutingTable

FILENAME = "foobar.html"


class StaticFileRoutingTableTest(unittest.TestCase):

    def setUp(self):
        self.routing_table = StaticFileRoutingTable(os.path.dirname(test_data.test_static.__file__))

    def test_whenGettingRouteRegistrations_thenReturnAnHttpGetRouteWhenItMatchesAFile(self):
        registrations = self.routing_table.get_route_registrations(FILENAME)

        self.assertEqual(1, len(registrations))

    def test_givenInexistentFile_whenGettingRouteRegistrations_thenReturnEmptyList(self):
        routes = self.routing_table.get_route_registrations("inexistent.html")

        self.assertEqual(0, len(routes))

    def test_givenDisallowedFileExtension_whenGettingRouteRegistrations_thenReturnEmptyList(self):
        self.routing_table = StaticFileRoutingTable(os.path.dirname(test_data.test_static.__file__),
                                                    allowed_extensions=[".txt"])

        routes = self.routing_table.get_route_registrations(FILENAME)

        self.assertEqual(0, len(routes))

    def test_givenDirectory_whenGettingRouteRegistrations_thenReturnEmptyList(self):
        routes = self.routing_table.get_route_registrations("")

        self.assertEqual(0, len(routes))



from typing import Iterable, List

from jivago.inject.service_locator import ServiceLocator
from jivago.lang.stream import Stream
from jivago.serialization.deserializer import Deserializer
from jivago.wsgi.invocation.route_handler import RouteHandler
from jivago.wsgi.methods import OPTIONS
from jivago.wsgi.request.request import Request
from jivago.wsgi.routing.cors.cors_request_handler_factory import CorsRequestHandlerFactory
from jivago.wsgi.routing.exception.method_not_allowed_exception import MethodNotAllowedException
from jivago.wsgi.routing.exception.unknown_path_exception import UnknownPathException
from jivago.wsgi.routing.routing_rule import RoutingRule


class RouteHandlerFactory(object):

    def __init__(self, service_locator: ServiceLocator,
                 deserializer: Deserializer,
                 routing_rules: List[RoutingRule],
                 cors_handler_factory: CorsRequestHandlerFactory):

        self.cors_handler_factory = cors_handler_factory
        self.routing_rules = routing_rules
        self.deserializer = deserializer
        self.service_locator = service_locator

    def create_route_handlers(self, request: Request) -> Iterable[RouteHandler]:
        route_registrations = Stream(self.routing_rules) \
            .map(lambda rule: rule.get_route_registrations(request.path)) \
            .flat().toList()

        routable_http_methods = Stream(route_registrations).map(lambda route: route.http_method).toSet()

        if len(route_registrations) == 0:
            raise UnknownPathException(request.path)

        if self.is_cors_request(request) and OPTIONS not in routable_http_methods:
            return Stream.of(self.cors_handler_factory.create_cors_preflight_handler(request.path))

        if request.method_annotation not in routable_http_methods:
            raise MethodNotAllowedException()

        return Stream(self.routing_rules) \
            .map(lambda rule: rule.create_route_handlers(request, self.service_locator, self.deserializer)) \
            .flat() \
            .map(lambda route_handler: self.cors_handler_factory.apply_cors_rules(request.path, route_handler))

    def is_cors_request(self, request: Request) -> bool:
        return request.method_annotation == OPTIONS



from typing import Generic, TypeVar

T = TypeVar('T')


class QueryParam(Generic[T]):
    pass


class OptionalQueryParam(Generic[T]):
    pass


class PathParam(Generic[T]):
    pass


class RequestBody(dict):
    pass



import types

from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Override
from jivago.wsgi.invocation.incorrect_resource_parameters_exception import IncorrectResourceParametersException
from jivago.wsgi.invocation.missing_route_invocation_argument import MissingRouteInvocationArgument
from jivago.wsgi.invocation.parameter_selection.parameter_selector_chain import ParameterSelectorChain
from jivago.wsgi.invocation.route_handler import RouteHandler
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response
from jivago.wsgi.routing.route_registration import RouteRegistration


class ResourceInvoker(RouteHandler):

    def __init__(self, route_registration: RouteRegistration,
                 service_locator: ServiceLocator,
                 parameter_selector_chain: ParameterSelectorChain):
        
        self.parameter_selector_chain = parameter_selector_chain
        self.route_registration = route_registration
        self.service_locator = service_locator

    @Override
    def invoke(self, request: Request) -> Response:
        try:
            method = types.MethodType(self.route_registration.routeFunction, self.get_resource_instance())
            parameters = self.parameter_selector_chain.get_parameters(request, method)
            function_return = method(*parameters)
        except MissingRouteInvocationArgument:
            raise IncorrectResourceParametersException()

        if isinstance(function_return, Response):
            return function_return
        return Response(200, {}, function_return)

    def get_resource_instance(self):
        return self.service_locator.get(self.route_registration.resourceClass) if isinstance(
            self.route_registration.resourceClass, type) else self.route_registration.resourceClass



from jivago.event.dispatch.jit_message_dispatcher_class import JitMessageDispatcherClass
from jivago.inject.service_locator import ServiceLocator


class MessageDispatcherRunnable(JitMessageDispatcherClass):

    def __init__(self, message_name: str, clazz: type, service_locator: ServiceLocator):
        super().__init__(message_name, clazz, clazz.run, service_locator)



from jivago.lang.registry import Annotation


class HttpMethod(Annotation):

    def __init__(self, name_str: str):
        super().__init__()
        self.name_str = name_str

    def __repr__(self):
        return self.name_str


GET = HttpMethod("GET")
POST = HttpMethod("POST")
DELETE = HttpMethod("DELETE")
PUT = HttpMethod("PUT")
PATCH = HttpMethod("PATCH")
OPTIONS = HttpMethod("OPTIONS")

http_methods = [GET, POST, DELETE, PUT, PATCH, OPTIONS]
method_strings = {'GET': GET, 'POST': POST, 'DELETE': DELETE, 'PUT': PUT, 'PATCH': PATCH, 'OPTIONS': OPTIONS}


def to_method(method_name: str) -> HttpMethod:
    return method_strings[method_name]



import unittest

from jivago.event.dispatch.message_dispatcher_function import MessageDispatcherFunction

PAYLOAD = object()


class MessageDispatcherFunctionTest(unittest.TestCase):

    def test_whenDispatching_thenInvokeFunction(self):
        dispatcher = MessageDispatcherFunction("foobar",dispatch)

        result = dispatcher.handle(PAYLOAD)

        self.assertEqual(PAYLOAD, result)


def dispatch(payload):
    return payload



from datetime import datetime

from croniter import croniter

from jivago.lang.annotations import Override
from jivago.scheduling.schedule import Schedule


class CronSchedule(Schedule):

    def __init__(self, cron_string: str, start_time: datetime):
        if start_time:
            self.iterator = croniter(cron_string, start_time=start_time.timestamp())
        else:
            self.iterator = croniter(cron_string)

    @Override
    def next_start_time(self) -> datetime:
        return self.iterator.get_next(ret_type=datetime)



from typing import NamedTuple, Type

from jivago.lang.annotations import Override
from jivago.serialization.deserialization_strategy import DeserializationStrategy


class NamedTupleDeserializationStrategy(DeserializationStrategy):

    def __init__(self, deserializer: "Deserializer"):
        self.deserializer = deserializer

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return hasattr(declared_type, "__bases__") \
               and tuple in declared_type.__bases__ \
               and hasattr(declared_type, "_field_types")

    @Override
    def deserialize(self, obj, declared_type: Type[NamedTuple]) -> NamedTuple:
        parameters = {}
        for name, clazz in declared_type._field_types.items():
            parameters[name] = self.deserializer.deserialize(obj[name], clazz)

        return declared_type(**parameters)



from jivago.config.properties.application_properties import ApplicationProperties


class ApplicationConfigLoader(object):

    def matches(self, filepath: str) -> bool:
        raise NotImplementedError

    def read(self, filepath: str) -> ApplicationProperties:
        raise NotImplementedError



from typing import Union, Callable


class Registration(object):
    def __init__(self, registered: Union[type, Callable], *, arguments=None):
        self.registered = registered
        self.arguments = arguments if arguments is not None else {}

    def is_in_package(self, package_name: str) -> bool:
        return self.registered.__module__.startswith(package_name)

    def is_class_registration(self):
        return isinstance(self.registered, type)

    def is_function_registration(self):
        return (not self.is_class_registration()
                and isinstance(self.registered, Callable)
                and not self.is_method_registration())

    def is_method_registration(self):
        return (not self.is_class_registration() and isinstance(self.registered, Callable)
                and self.registered.__code__.co_argcount > 0
                and self.registered.__code__.co_varnames[0] == "self")



from typing import List, Type

from jivago.config.production_jivago_context import ProductionJivagoContext
from jivago.config.router.cors_rule import CorsRule
from jivago.config.router.router_builder import RouterBuilder
from jivago.lang.annotations import Override
from jivago.lang.registry import Registry
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.system_filters.error_handling.debug_exception_filter import DebugExceptionFilter
from jivago.wsgi.filter.system_filters.error_handling.unknown_exception_filter import UnknownExceptionFilter


class DebugJivagoContext(ProductionJivagoContext):

    def __init__(self, root_package: "Module", registry: Registry, banner: bool = True):
        super().__init__(root_package, registry, banner=banner)

    def create_router_config(self) -> RouterBuilder:
        return super().create_router_config().add_rule(CorsRule("/", {"Access-Control-Allow-Origin": '*',
                                                                      'Access-Control-Allow-Headers': '*',
                                                                      'Access-Control-Allow-Methods': '*'}))

    @Override
    def get_default_filters(self) -> List[Type[Filter]]:
        production_filters = super().get_default_filters()
        production_filters.remove(UnknownExceptionFilter)
        return [DebugExceptionFilter] + production_filters



from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class Filter(object):

    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        raise NotImplementedError



from jivago.lang.annotations import Override
from jivago.wsgi.filter.system_filters.error_handling.exception_mapper import ExceptionMapper
from jivago.wsgi.request.response import Response
from jivago.wsgi.routing.cors.no_matching_cors_rule_exception import NoMatchingCorsRuleException


class NoMatchingCorsRuleExceptionMapper(ExceptionMapper):

    @Override
    def handles(self, exception: Exception) -> bool:
        return isinstance(exception, NoMatchingCorsRuleException)

    @Override
    def create_response(self, exception: Exception) -> Response:
        return Response(400, {}, {"message": "Missing CORS rule."})



from typing import List

from jivago.lang.registry import Annotation
from jivago.lang.stream import Stream


class ScopeCache(object):

    def __init__(self, scope: Annotation, scoped_components: List[type]):
        self.scope = scope
        self.scoped_components = Stream(scoped_components).map(lambda clazz: (clazz, None)).toDict()

    def handles_component(self, component: type) -> bool:
        return component in self.scoped_components.keys()

    def is_stored(self, component: type) -> bool:
        return self.scoped_components[component] is not None

    def get(self, component: type) -> object:
        stored_component = self.scoped_components.get(component)
        if stored_component is None:
            raise UninstantiatedObjectException(component)
        return stored_component

    def store(self, component: type, instance: object):
        if component not in self.scoped_components.keys():
            raise ComponentNotHandledByScopeException(component)
        self.scoped_components[component] = instance


class ComponentNotHandledByScopeException(Exception):
    pass


class UninstantiatedObjectException(Exception):
    pass



import inspect
from typing import Callable

from jivago.config.abstract_binder import AbstractBinder
from jivago.inject.annotation import Provider
from jivago.inject.exception.undefined_return_provider_function import UndefinedReturnProviderFunction
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Override
from jivago.lang.registry import Registry
from jivago.lang.stream import Stream


class ProviderBinder(AbstractBinder):

    def __init__(self, root_package: str, registry: Registry):
        self.rootPackage = root_package
        self.registry = registry

    @Override
    def bind(self, service_locator: ServiceLocator):
        providers = self.registry.get_annotated_in_package(Provider, self.rootPackage)
        Stream(providers).map(lambda r: r.registered) \
            .forEach(lambda p: service_locator.bind(_function_return_type(p), p))


def _function_return_type(fun: Callable) -> type:
    return_type = inspect.signature(fun).return_annotation
    if return_type == inspect._empty:
        raise UndefinedReturnProviderFunction()
    return return_type



import threading
import time
from datetime import datetime

from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Override
from jivago.lang.runnable import Runnable
from jivago.scheduling.schedule import Schedule


class ScheduledTaskRunner(Runnable):

    def __init__(self, runner_class: type, schedule: Schedule, service_locator: ServiceLocator):
        self.service_locator = service_locator
        self.runner_class = runner_class
        self.schedule = schedule
        self.thread_stop_event = threading.Event()
        self.thread = threading.Thread(target=self.run, daemon=True)
        self.run_lock = threading.Lock()

    @Override
    def run(self):
        while not self.thread_stop_event.is_set():
            sleep_time = self.schedule.next_start_time() - datetime.utcnow()
            if sleep_time.total_seconds() > 0:
                time.sleep(sleep_time.total_seconds())
            self.run_lock.acquire()
            self.service_locator.get(self.runner_class).run()
            self.run_lock.release()

    def stop(self):
        self.thread_stop_event.set()
        self.run_lock.acquire()
        self.service_locator.get(self.runner_class).cleanup()
        self.run_lock.release()

    def start(self):
        self.thread.start()



import unittest

from jivago.wsgi.invocation.parameter_selection.query_parameter_selector import QueryParameterSelector
from jivago.wsgi.invocation.parameters import QueryParam


class QueryParameterSelectorTest(unittest.TestCase):

    def setUp(self):
        self.selector = QueryParameterSelector()

    def test_whenMatchingFunction_thenMatchQueryParameterTypes(self):
        self.assertTrue(self.selector.matches(QueryParam[str]))
        self.assertFalse(self.selector.matches(str))



from datetime import datetime, date
from typing import Type

import dateutil.parser

from jivago.lang.annotations import Override
from jivago.serialization.deserialization_strategy import DeserializationStrategy, T


class DatetimeDeserializationStrategy(DeserializationStrategy):

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return declared_type == datetime or declared_type == date

    @Override
    def deserialize(self, obj, declared_type: Type[T]) -> T:
        parsed_date = dateutil.parser.parse(obj)
        if declared_type == date:
            return parsed_date.date()
        return parsed_date



import unittest
from unittest import mock

from jivago.event.dispatch.message_dispatcher import MessageDispatcher
from jivago.event.synchronous_event_bus import SynchronousEventBus

MESSAGE_NAME = "message_name"

PAYLOAD = object()


class SynchronousEventBusTest(unittest.TestCase):

    def setUp(self):
        self.message_dispatcher: MessageDispatcher = mock.create_autospec(MessageDispatcher)
        self.message_bus = SynchronousEventBus([self.message_dispatcher])
        self.message_dispatcher.can_handle.return_value = True

    def test_whenEmittingEvent_thenInvokeMatchingMessageDispatcher(self):
        self.message_bus.emit(MESSAGE_NAME, PAYLOAD)

        self.message_dispatcher.handle.assert_called_with(PAYLOAD)

    def test_whenEmittingEvent_thenReturnCollectionOfResponses(self):
        self.message_dispatcher.handle.return_value = PAYLOAD

        responses = self.message_bus.emit(MESSAGE_NAME, PAYLOAD)

        self.assertEqual(1, len(responses))
        self.assertEqual(PAYLOAD, responses[0])



import logging

from jivago.inject.service_locator import ServiceLocator
from jivago.wsgi.filter.filter_chain_factory import FilterChainFactory
from jivago.wsgi.request.http_status_code_resolver import HttpStatusCodeResolver
from jivago.wsgi.request.request_factory import RequestFactory
from jivago.wsgi.request.response import Response


class Router(object):
    LOGGER = logging.getLogger("Jivago").getChild("Router")

    def __init__(self, service_locator: ServiceLocator,
                 request_factory: RequestFactory,
                 filter_chain_factory: FilterChainFactory):
        self.filter_chain_factory = filter_chain_factory
        self.request_factory = request_factory
        self.serviceLocator = service_locator
        self.http_status_resolver = HttpStatusCodeResolver()

    def route(self, env, start_response):
        request = self.request_factory.build_request(env)

        filter_chain = self.filter_chain_factory.create_filter_chain(request)

        response = Response.empty()

        filter_chain.doFilter(request, response)

        start_response(self.http_status_resolver.get_status_code(response.status),
                       [x for x in response.headers.items()])
        if isinstance(response.body, str):
            return [response.body.encode('utf-8')]
        else:
            return [response.body]



from jivago.lang.annotations import Override
from jivago.wsgi.invocation.missing_route_invocation_argument import MissingRouteInvocationArgument
from jivago.wsgi.invocation.parameter_selection.parameter_selector import ParameterSelector
from jivago.wsgi.invocation.parameters import PathParam
from jivago.wsgi.request.request import Request
from jivago.wsgi.routing.path_parameter_parser import PathParameterParser

ALLOWED_PATH_PARAMETER_TYPES = (str, int, float)


class PathParameterSelector(ParameterSelector):

    def __init__(self, path_parameter_parser: PathParameterParser):
        self.path_parameter_parser = path_parameter_parser

    @Override
    def matches(self, paramater_declaration: type) -> bool:
        return hasattr(paramater_declaration, '__origin__') and paramater_declaration.__origin__ == PathParam

    @Override
    def format_parameter(self, parameter_name: str, parameter_declaration: type, request: Request) -> object:
        parameter_type = parameter_declaration.__args__[0]
        if parameter_type in ALLOWED_PATH_PARAMETER_TYPES:
            try:
                return parameter_type(self.path_parameter_parser.parse_path_parameters(request.path)[parameter_name])
            except (KeyError, ValueError):
                raise MissingRouteInvocationArgument(parameter_name)
        raise MissingRouteInvocationArgument(parameter_name)



import json
from typing import Optional

from jivago.lang.annotations import Override
from jivago.lang.nullable import Nullable
from jivago.lang.stream import Stream
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class JsonSerializationFilter(Filter):

    @Override
    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        if is_application_json(request.headers['Content-Type']) and len(request.body) > 0:
            request.body = json.loads(request.body)

        chain.doFilter(request, response)

        if isinstance(response.body, dict) or isinstance(response.body, list):
            response.body = json.dumps(response.body)
            response.headers['Content-Type'] = 'application/json'


def is_application_json(content_type: Optional[str]) -> bool:
    return Nullable(content_type) \
        .map(lambda x: "application/json" in x) \
        .orElse(False)



from typing import List

from jivago.lang.annotations import Override, Inject
from jivago.lang.stream import Stream
from jivago.wsgi.filter.system_filters.error_handling.exception_mapper import ExceptionMapper
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class ApplicationExceptionFilter(Filter):

    @Inject
    def __init__(self, exception_mappers: List[ExceptionMapper]):
        self.exception_mappers = exception_mappers

    @Override
    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        try:
            chain.doFilter(request, response)
        except Exception as e:
            exception_mapper = Stream(self.exception_mappers).firstMatch(lambda mapper: mapper.handles(e))
            if exception_mapper.isPresent():
                response.copy(exception_mapper.get().create_response(e))
            else:
                raise e



import unittest
from unittest import mock

from jivago.wsgi.invocation.rewrite.path_rewriting_route_handler_decorator import PathRewritingRouteHandlerDecorator
from jivago.wsgi.invocation.route_handler import RouteHandler
from test_utils.request_builder import RequestBuilder

NEW_PATH = "/new-path"


class PathRewritingRouteHandlerDecoratorTest(unittest.TestCase):

    def setUp(self):
        self.decorated_route_handler: RouteHandler = mock.create_autospec(RouteHandler)
        self.route_handler = PathRewritingRouteHandlerDecorator(self.decorated_route_handler, NEW_PATH)

    def test_whenInvoking_thenPassModifiedRequestWithModifiedPath(self):
        request = RequestBuilder().path("/old-path").build()

        self.route_handler.invoke(request)

        _, args, _ = self.decorated_route_handler.invoke.mock_calls[0]
        gotten_request = args[0]

        self.assertEqual(NEW_PATH, gotten_request.path)



class MessageDispatcher(object):

    def __init__(self, message_name: str):
        self.message_name = message_name

    def can_handle(self, message_name: str) -> bool:
        return self.message_name == message_name

    async def handle(self, payload: object):
        raise NotImplementedError



import inspect
from typing import Type

from jivago.lang.annotations import Override, Serializable
from jivago.lang.registry import Registry
from jivago.lang.stream import Stream
from jivago.serialization.deserialization_strategy import DeserializationStrategy, T


class RegisteredSerializableTypeDeserializationStrategy(DeserializationStrategy):

    def __init__(self, registry: Registry, deserializer: "Deserializer"):
        self.deserializer = deserializer
        self.registry = registry

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return self.registry.is_annotated(declared_type, Serializable)

    @Override
    def deserialize(self, obj, declared_type: Type[T]) -> T:
        if _has_defined_initializer(declared_type):
            initializer_signature = inspect.signature(declared_type)
            return declared_type(**Stream(initializer_signature.parameters.items()) \
                                 .map(lambda name, parameter:
                                      (name, self.deserializer.deserialize(obj.get(name), parameter.annotation))) \
                                 .toDict())

        else:
            instance = object.__new__(declared_type)
            attributes = declared_type.__annotations__
            Stream(attributes.items()) \
                .map(lambda name, attribute_type:
                     (name, self.deserializer.deserialize(obj.get(name), attribute_type))) \
                .forEach(lambda name, attribute: instance.__setattr__(name, attribute))

            return instance


def _has_defined_initializer(type: type) -> bool:
    return type.__init__ != object.__init__



from typing import List

from jivago.config.properties.application_config_loader import ApplicationConfigLoader
from jivago.config.properties.application_properties import ApplicationProperties
from jivago.config.properties.exception import ConfigurationException
from jivago.lang.annotations import Override
from jivago.lang.stream import Stream


class GlobalConfigLoader(ApplicationConfigLoader):

    def __init__(self, application_config_loaders: List[ApplicationConfigLoader]):
        self.application_config_loaders = application_config_loaders

    @Override
    def matches(self, filepath: str) -> bool:
        return Stream(self.application_config_loaders) \
            .anyMatch(lambda loader: loader.matches(filepath))

    @Override
    def read(self, filepath: str) -> ApplicationProperties:
        return Stream(self.application_config_loaders) \
            .firstMatch(lambda loader: loader.matches(filepath)) \
            .map(lambda x: x.read(filepath)) \
            .orElseThrow(ConfigurationException())



from typing import Type, TypeVar

T = TypeVar('T')


class DeserializationStrategy(object):

    def can_handle_deserialization(self, declared_type: type) -> bool:
        raise NotImplementedError

    def deserialize(self, obj, declared_type: Type[T]) -> T:
        raise NotImplementedError



import inspect
from typing import Callable

from jivago.lang.nullable import Nullable
from jivago.lang.stream import Stream
from jivago.serialization.deserializer import Deserializer
from jivago.wsgi.invocation.missing_route_invocation_argument import MissingRouteInvocationArgument
from jivago.wsgi.invocation.parameter_selection.dictionary_parameter_selector import DictionaryParameterSelector
from jivago.wsgi.invocation.parameter_selection.headers_parameter_selector import HeadersParameterSelector
from jivago.wsgi.invocation.parameter_selection.optional_query_parameter_selector import OptionalQueryParameterSelector
from jivago.wsgi.invocation.parameter_selection.parameter_selector import ParameterSelector
from jivago.wsgi.invocation.parameter_selection.path_parameter_selector import PathParameterSelector
from jivago.wsgi.invocation.parameter_selection.query_parameter_selector import QueryParameterSelector
from jivago.wsgi.invocation.parameter_selection.raw_request_parameter_selector import RawRequestParameterSelector
from jivago.wsgi.invocation.parameter_selection.serialized_parameter_selector import SerializedParameterSelector
from jivago.wsgi.request.request import Request
from jivago.wsgi.routing.route_registration import RouteRegistration


class ParameterSelectorChain(object):

    def __init__(self, route: RouteRegistration, deserializer: Deserializer):
        self.parameter_selectors = [
            RawRequestParameterSelector(),
            HeadersParameterSelector(),
            DictionaryParameterSelector(),
            QueryParameterSelector(),
            OptionalQueryParameterSelector(),
            PathParameterSelector(route),
            SerializedParameterSelector(deserializer),
        ]

    def get_parameters(self, request: Request, method: Callable) -> list:
        parameters = []
        for name, parameter in inspect.signature(method).parameters.items():
            parameter_type = parameter._annotation
            selector: Nullable[ParameterSelector] = Stream(self.parameter_selectors) \
                .filter(lambda s: s.matches(parameter_type)) \
                .first()

            if selector.isPresent():
                parameters.append(selector.get().format_parameter(name, parameter_type, request))
            else:
                raise MissingRouteInvocationArgument(name)

        return parameters



from jivago.config.router.router_config_rule import RouterConfigRule
from jivago.wsgi.request.headers import Headers


class CorsRule(RouterConfigRule):

    def __init__(self, path: str, cors_headers: dict):
        self.path = path
        self.cors_headers = Headers(cors_headers)

    def matches(self, path: str) -> bool:
        return path.startswith(self.path)

    def takes_precedence_over(self, other: "CorsRule") -> bool:
        return len(self.path) >= len(other.path)

    def inject_headers(self, headers: Headers) -> None:
        for key, value in self.cors_headers.items():
            headers[key] = value



import logging
import traceback

from jivago.lang.annotations import Override
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class DebugExceptionFilter(Filter):
    LOGGER = logging.getLogger("DebugExceptionFilter")

    @Override
    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        try:
            chain.doFilter(request, response)
        except Exception as e:
            response.status = 500
            response.body = "{}\n{}".format(e.__class__.__name__, traceback.format_exc())
            self.LOGGER.error(traceback.format_exc())



from typing import List

from jivago.wsgi.routing.route_registration import RouteRegistration


class RoutingTable(object):

    def get_route_registrations(self, path: str) -> List[RouteRegistration]:
        raise NotImplementedError



import unittest

from jivago.event.config.annotations import EventHandler, EventHandlerClass
from jivago.event.config.reflective_event_bus_initializer import ReflectiveEventBusInitializer
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.registry import Registry
from jivago.lang.runnable import Runnable
from jivago.lang.stream import Stream


class ReflectiveMessageBusInitializerTest(unittest.TestCase):

    def setUp(self):
        self.event_bus_initializer = ReflectiveEventBusInitializer(ServiceLocator(), Registry(), "")

    def test_whenInitializing_thenRegisterHandlerFunctions(self):
        bus = self.event_bus_initializer.create_message_bus()

        self.assertMessageNameInBus("handler_function", bus)

    def test_whenInitializing_thenRegisterJitHandlerClass(self):
        bus = self.event_bus_initializer.create_message_bus()

        self.assertMessageNameInBus("handler_method", bus)

    def test_whenInitializing_thenRegisterRunnable(self):
        bus = self.event_bus_initializer.create_message_bus()

        self.assertMessageNameInBus("handler_runnable", bus)

    def assertMessageNameInBus(self, message: str, bus):
        self.assertTrue(message in Stream(bus.message_handlers).map(lambda x: x.message_name).toTuple())


@EventHandler(event_name="handler_function")
def handler_function():
    return "ok"


@EventHandlerClass
class HandlerClass(object):

    @EventHandler(event_name="handler_method")
    def handler_method(self):
        return "method"


@EventHandler(event_name="handler_runnable")
class HandlerRunnable(Runnable):

    def run(self):
        pass



from jivago.config.router.filtering.annotation import RequestFilter
from jivago.config.router.filtering.filtering_rule import FilteringRule
from jivago.lang.registry import Registry
from jivago.lang.stream import Stream


class AutoDiscoveringFilteringRule(FilteringRule):

    def __init__(self, url_pattern: str, registry: Registry, root_package_name: str, regex_pattern: str = None):
        registrations = registry.get_annotated_in_package(RequestFilter, root_package_name)

        super().__init__(url_pattern, Stream(registrations).map(lambda r: r.registered).toList(), regex_pattern)



import unittest
from unittest import mock

from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.filter.system_filters.error_handling.application_exception_filter import ApplicationExceptionFilter
from jivago.wsgi.filter.system_filters.error_handling.exception_mapper import ExceptionMapper
from jivago.wsgi.request.response import Response


class ApplicationExceptionFilterTest(unittest.TestCase):

    def setUp(self):
        self.exceptionMapperMock: ExceptionMapper = mock.create_autospec(ExceptionMapper)
        self.filterChainMock: FilterChain = mock.create_autospec(FilterChain)
        self.filter = ApplicationExceptionFilter([self.exceptionMapperMock])

    def test_givenNoMatchingExceptionMapper_whenHandlingException_thenRaiseTheSameException(self):
        self.exceptionMapperMock.handles.return_value = False
        thrown_exception = AnException()
        self.filterChainMock.doFilter.side_effect = thrown_exception

        try:
            self.filter.doFilter(None, None, self.filterChainMock)
            self.fail()
        except AnException as e:
            self.assertEqual(thrown_exception, e)

    def test_givenMatchingExceptionMapper_whenHandlingException_thenReturnTheCreatedHttpResponse(self):
        created_response = Response(200, {"foo": "bar"}, "body")
        self.exceptionMapperMock.handles.return_value = True
        self.exceptionMapperMock.create_response.return_value = created_response
        self.filterChainMock.doFilter.side_effect = AnException()

        returned_response = Response.empty()
        self.filter.doFilter(None, returned_response, self.filterChainMock)

        self.assertEqual(created_response.headers, returned_response.headers)
        self.assertEqual(created_response.body, returned_response.body)
        self.assertEqual(created_response.status, returned_response.status)


class AnException(Exception):
    pass



import unittest

from jivago.wsgi.request.request_factory import RequestFactory


class RequestFactoryTest(unittest.TestCase):

    def setUp(self):
        self.request_factory = RequestFactory()

    def test_whenBuildingRequest_thenFormatCustomHttpHeadersUsingCamelCase(self):
        request = self.request_factory.build_request(WSGI_ENV)

        self.assertTrue(("Foo", "bar") in request.headers.items())

    def test_givenCustomHeaderWithUnderscores_whenBuildingRequest_thenReplaceUnderscoresWithHyphens(self):
        request = self.request_factory.build_request(WSGI_ENV)

        self.assertTrue(("Foo-Bar", "bar") in request.headers.items())

    def test_whenBuildingRequest_thenPopulateRequiredHeaderFields(self):
        request = self.request_factory.build_request(WSGI_ENV)

        self.assertEqual("GET", request.method)
        self.assertEqual("/", request.path)
        self.assertEqual("", request.queryString)

    def test_givenCustomHeaderStartingWithP_whenBuildingRequest_thenDoNotRemoveTheLeadingP(self):
        request = self.request_factory.build_request(WSGI_ENV)

        self.assertEqual("postman", request.headers['Postman-Token'])

    def test_givenJsonPost_whenBuildingRequest_thenAddContentTypeHeader(self):
        request = self.request_factory.build_request(POST_WSGI_ENV)

        self.assertEqual(b'{\n"message" : "string"\n}', request.body)
        self.assertEqual("application/json", request.headers['Content-Type'])


class DummyWsgiReadStream(object):
    def __init__(self, content=b""):
        self.content = content

    def read(self, *args):
        return self.content


WSGI_ENV = {
    'HTTP_FOO': 'bar',  # custom header
    'HTTP_FOO_BAR': 'bar',  # custom header
    'wsgi.version': (1, 0), 'wsgi.url_scheme': 'http', 'wsgi.input': DummyWsgiReadStream(), 'wsgi.errors': lambda x: x,
    'wsgi.multithread': False, 'wsgi.multiprocess': False, 'wsgi.run_once': False,
    'werkzeug.server.shutdown': lambda x: x,
    'SERVER_SOFTWARE': 'Werkzeug/0.14.1',
    'REQUEST_METHOD': 'GET',
    'SCRIPT_NAME': '',
    'PATH_INFO': '/',
    'QUERY_STRING': '',
    'HTTP_POSTMAN_TOKEN': "postman",
    'REMOTE_ADDR': '127.0.0.1', 'REMOTE_PORT': 58576,
    'SERVER_NAME': '127.0.0.1', 'SERVER_PORT': '4000',
    'SERVER_PROTOCOL': 'HTTP/1.1', 'HTTP_HOST': 'localhost:4000',
    'HTTP_CONNECTION': 'keep-alive',
    'HTTP_CACHE_CONTROL': 'no-cache',
    'HTTP_USER_AGENT': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/70.0.3538.77 Chrome/70.0.3538.77 Safari/537.36',
    'HTTP_ACCEPT': '*/*', 'HTTP_ACCEPT_ENCODING': 'gzip, deflate, br',
    'HTTP_ACCEPT_LANGUAGE': 'en-US,en;q=0.9,fr-CA;q=0.8,fr;q=0.7'}

POST_WSGI_ENV = {
    'CONTENT_TYPE': 'application/json',
    'wsgi.version': (1, 0), 'wsgi.url_scheme': 'http',
    'wsgi.input': DummyWsgiReadStream(b'{\n"message" : "string"\n}'),
    'wsgi.errors': lambda x: x,
    'wsgi.multithread': False,
    'wsgi.multiprocess': False,
    'wsgi.run_once': False,
    'werkzeug.server.shutdown': lambda x: x,
    'SERVER_SOFTWARE': 'Werkzeug/0.14.1', 'REQUEST_METHOD': 'POST',
    'SCRIPT_NAME': '', 'PATH_INFO': '/', 'QUERY_STRING': '', 'REMOTE_ADDR': '127.0.0.1',
    'REMOTE_PORT': 47858, 'SERVER_NAME': '127.0.0.1', 'SERVER_PORT': '4000',
    'SERVER_PROTOCOL': 'HTTP/1.1', 'HTTP_HOST': 'localhost:4000',
    'HTTP_CONNECTION': 'keep-alive', 'CONTENT_LENGTH': '24',
    'HTTP_USER_AGENT': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/70.0.3538.77 Chrome/70.0.3538.77 Safari/537.36',
    'HTTP_CACHE_CONTROL': 'no-cache',
    'HTTP_ACCEPT': '*/*', 'HTTP_ACCEPT_ENCODING': 'gzip, deflate, br',
    'HTTP_ACCEPT_LANGUAGE': 'en-US,en;q=0.9,fr-CA;q=0.8,fr;q=0.7'}



import unittest
from unittest import mock

from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.invocation.incorrect_resource_parameters_exception import IncorrectResourceParametersException
from jivago.wsgi.invocation.resource_invoker import ResourceInvoker
from jivago.wsgi.invocation.route_handler_factory import RouteHandlerFactory
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response

A_REQUEST = Request('GET', "/path", {}, "", "")
A_RESPONSE = Response.empty()


class FilterChainTest(unittest.TestCase):

    def setUp(self):
        self.resourceInvokerFactoryMock: RouteHandlerFactory = mock.create_autospec(RouteHandlerFactory)
        self.resourceInvokerMock: ResourceInvoker = mock.create_autospec(ResourceInvoker)
        self.filterMock: Filter = mock.create_autospec(Filter)
        self.secondFilterMock: Filter = mock.create_autospec(Filter)
        self.filterChain = FilterChain([self.filterMock, self.secondFilterMock], self.resourceInvokerFactoryMock)
        self.resourceInvokerFactoryMock.create_route_handlers.return_value = [self.resourceInvokerMock]

    def test_givenEmptyFilterChain_whenApplyingFilterChain_thenInvokeResourceInvocator(self):
        self.filterChain = FilterChain([], self.resourceInvokerFactoryMock)

        self.filterChain.doFilter(A_REQUEST, A_RESPONSE)

        self.resourceInvokerFactoryMock.create_route_handlers.assert_called_with(A_REQUEST)

    def test_givenNonEmptyFilterChain_whenGettingNextChain_thenReturnANewChainWithOneFewerFilter(self):
        next_chain = self.filterChain.getNextChain()

        self.assertEqual(1, len(next_chain.filters))
        self.assertEqual(self.secondFilterMock, next_chain.filters[0])

    def test_whenApplyingFilterChain_thenInvokeNextFilterInTheChain(self):
        self.filterChain.doFilter(A_REQUEST, A_RESPONSE)

        self.filterMock.doFilter.assert_called_once()
        self.secondFilterMock.doFilter.assert_not_called()

    def test_givenExceptionInOneOfTheResourceInvokers_whenInvoking_thenSilenceTheException(self):
        self.second_resource_invoker_mock: ResourceInvoker = mock.create_autospec(ResourceInvoker)
        self.second_resource_invoker_mock.invoke.side_effect = IncorrectResourceParametersException()
        self.resourceInvokerFactoryMock.create_route_handlers.return_value = [self.second_resource_invoker_mock, self.resourceInvokerMock]
        self.filterChain.filters = []

        self.filterChain.doFilter(A_REQUEST, A_RESPONSE)

        self.resourceInvokerMock.invoke.assert_called_with(A_REQUEST)




from jivago.config.abstract_binder import AbstractBinder
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Override
from jivago.lang.registry import Registry, Annotation


class AnnotatedClassBinder(AbstractBinder):

    def __init__(self, root_package: str, registry: Registry, annotation: Annotation):
        self.annotation = annotation
        self.rootPackage = root_package
        self.registry = registry

    @Override
    def bind(self, service_locator: ServiceLocator):
        for registration in self.registry.get_annotated_in_package(self.annotation, self.rootPackage):
            component = registration.registered
            service_locator.bind(component, component)



import unittest

from jivago.wsgi.methods import GET
from jivago.wsgi.routing.exception.method_not_allowed_exception import MethodNotAllowedException
from jivago.wsgi.routing.exception.unknown_path_exception import UnknownPathException
from jivago.wsgi.routing.table.tree_routing_table import TreeRoutingTable

PATH = "/foobar/hello"

HTTP_PRIMITIVE = GET


class TreeRoutingTableTest(unittest.TestCase):

    def setUp(self):
        self.routing_table = TreeRoutingTable()
        self.routing_table.register_route(HTTP_PRIMITIVE, PATH, A_Resource, A_Resource.get_hello)

    def test_whenRegisteringRoute_thenRouteRegistrationIsSaved(self):
        route_registrations = self.routing_table.get_route_registrations(PATH)

        self.assertEqual(["foobar", "hello"], route_registrations[0].registeredPath)
        self.assertEqual(A_Resource, route_registrations[0].resourceClass)
        self.assertEqual(A_Resource.get_hello, route_registrations[0].routeFunction)


class A_Resource(object):

    def get_hello(self):
        return "hello"



import unittest

from jivago.serialization.object_mapper import ObjectMapper


class ObjectMapperTest(unittest.TestCase):

    def setUp(self):
        self.object_mapper = ObjectMapper()

    def test_givenUnregisteredDtoClass_whenDeserializing_thenIgnoreClassRegistrationStatus(self):
        dto = self.object_mapper.deserialize('{ "name": "foo" }', A_Dto)

        self.assertEqual("foo", dto.name)

    def test_givenUnregisteredDtoClass_whenSerializing_thenIgnoreClassRegistrationStatus(self):
        dto = A_Dto()
        dto.name = "foo"

        json_str = self.object_mapper.serialize(dto)

        self.assertEqual('{"name": "foo"}', json_str)


class A_Dto(object):
    name: str



from typing import List, Callable, Union

from jivago.lang.annotations import Override
from jivago.lang.stream import Stream
from jivago.wsgi.methods import HttpMethod
from jivago.wsgi.routing.exception.unknown_path_exception import UnknownPathException
from jivago.wsgi.routing.route_registration import RouteRegistration
from jivago.wsgi.routing.routing_table import RoutingTable
from jivago.wsgi.routing.table.path_util import split_path
from jivago.wsgi.routing.table.route_node import RouteNode


class TreeRoutingTable(RoutingTable):

    def __init__(self):
        self.root_node = RouteNode()

    def register_route(self, primitive: HttpMethod, path: str, resource_class: Union[type, object],
                       route_method: Callable):
        path = split_path(path)
        self.root_node.register_child(path, primitive, RouteRegistration(resource_class, route_method, path, primitive))

    @Override
    def get_route_registrations(self, path: str) -> List[RouteRegistration]:
        path_elements = split_path(path)
        try:
            route_node = self.root_node.explore(path_elements)
            return Stream(route_node.invocators.values()).flat().toList()
        except UnknownPathException:
            return []



from typing import Callable

from jivago.lang.registry import Annotation


def Override(fun: Callable) -> Callable:
    return fun


@Annotation
def Inject(fun: Callable) -> Callable:
    return fun


@Annotation
def Serializable(clazz: type) -> type:
    return clazz


@Annotation
def BackgroundWorker(clazz: type) -> type:
    return clazz



from concurrent.futures import ThreadPoolExecutor

from jivago.event.event_bus import EventBus
from jivago.event.synchronous_event_bus import SynchronousEventBus
from jivago.lang.annotations import Override


class AsyncEventBus(EventBus):

    def __init__(self, event_bus: SynchronousEventBus, pool_size=12):
        self.event_bus = event_bus
        self.thread_pool = ThreadPoolExecutor(max_workers=pool_size)

    @Override
    def emit(self, message_name: str, payload=None):
        self.thread_pool.submit(self.event_bus.emit, message_name, payload)



from jivago.lang.annotations import Override
from jivago.wsgi.invocation.route_handler import RouteHandler
from jivago.wsgi.request.headers import Headers
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response

ALLOW_ORIGIN_HEADER = 'Access-Control-Allow-Origin'


class CorsPreflightRequestHandler(RouteHandler):

    def __init__(self, cors_headers: Headers):
        self.cors_headers = cors_headers
        self.allowed_origin = self.cors_headers[
            ALLOW_ORIGIN_HEADER] if ALLOW_ORIGIN_HEADER in self.cors_headers else '*'

    @Override
    def invoke(self, request: Request) -> Response:
        origin = request.headers['Origin']
        if self.should_allow(origin):
            response = Response(200, self.cors_headers, "")
        else:
            response = Response(400, self.cors_headers, "")

        if 'Vary' not in response.headers and response.headers[ALLOW_ORIGIN_HEADER] != '*':
            response.headers['Vary'] = 'Origin'

        return response

    def should_allow(self, origin: str) -> bool:
        if self.allowed_origin == '*':
            return True
        return origin.startswith(self.allowed_origin)



from jivago.lang.annotations import Inject, Serializable
from jivago.lang.registry import Registry
from jivago.serialization.serializer import Serializer
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response

TYPES_TO_BE_SERIALIZED = (dict, list)


class BodySerializationFilter(Filter):

    @Inject
    def __init__(self, serializer: Serializer, registry: Registry):
        self.registry = registry
        self.serializer = serializer

    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        chain.doFilter(request, response)

        if self.registry.is_annotated(response.body.__class__, Serializable) or type(response.body) in TYPES_TO_BE_SERIALIZED:
            response.body = self.serializer.serialize(response.body)



STATUS_CODES = {
    100: "Continue",
    101: "Switching Protocols",
    102: "Processing",
    103: "Early Hints",

    200: "OK",
    201: "Created",
    202: "Accepted",
    203: "Non-Authoritative Information",
    204: "No Content",
    205: "Reset Content",
    206: "Partial Content",
    207: "Multi-Status",
    208: "Already Reported",
    226: "IM Used",

    300: "Multiple Choices",
    301: "Moved Permanently",
    302: "Found",
    303: "See Other",
    304: "Not Modified",
    305: "Use Proxy",
    306: "Switch Proxy",
    307: "Temporary Redirect",
    308: "Permanent Redirect",

    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Payload Too Large",
    414: "URI Too Long",
    415: "Unsupported Media Type",
    416: "Range Not Satisfiable",
    417: "Expectation Failed",
    418: "I'm a teapot",
    421: "Misdirected Request",
    422: "Unprocessable Entity",
    423: "Locked",
    424: "Failed Dependency",
    426: "Upgrade Required",
    428: "Precondition Required",
    429: "Too Many Requests",
    431: "Request Header Fields Too Large",
    451: "Unavailable For Legal Reasons",

    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Timeout",
    505: "HTTP Version Not Supported",
    506: "Variant Also Negotiates",
    507: "Insufficient Storage",
    508: "Loop Detected",
    510: "Not Extended",
    511: "Network Authentication Required"
}


class HttpStatusCodeResolver(object):

    def get_status_code(self, status: int) -> str:
        return f"{status} {STATUS_CODES[status]}" if status in STATUS_CODES else str(status)



from threading import Lock

from jivago.inject.service_locator import ServiceLocator
from jivago.lang.stream import Stream
from jivago.scheduling.schedule import Schedule
from jivago.scheduling.scheduled_task_runner import ScheduledTaskRunner


class TaskScheduler(object):

    def __init__(self, service_locator: ServiceLocator):
        self.service_locator = service_locator
        self.task_runners = []
        self.lock = Lock()

    def schedule_task(self, runnable_class: type, schedule: Schedule):
        with self.lock:
            task_runner = ScheduledTaskRunner(runnable_class, schedule, self.service_locator)
            self.task_runners.append(task_runner)
            task_runner.start()

    def stop(self):
        with self.lock:
            Stream(self.task_runners).forEach(lambda x: x.stop())



import unittest

from jivago.event.dispatch.message_dispatcher import MessageDispatcher

MESSAGE_NAME = "message_name"


class MessageDispatcherTest(unittest.TestCase):

    def setUp(self):
        self.dispatcher = MessageDispatcher(MESSAGE_NAME)

    def test_whenMatchingMessageName_thenCompareStrings(self):
        can_handle = self.dispatcher.can_handle(MESSAGE_NAME)

        self.assertTrue(can_handle)

    def test_givenDifferentMessageName_whenCheckingCanHandle_thenDispatcherCannotHandleMessage(self):
        can_handle = self.dispatcher.can_handle("wrong message")

        self.assertFalse(can_handle)



import unittest
from unittest import mock

from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.filter.system_filters.jivago_banner_filter import JivagoBannerFilter
from test_utils.request_builder import RequestBuilder
from test_utils.response_builder import ResponseBuilder


class JivagoBannerFilterTest(unittest.TestCase):

    def setUp(self):
        self.filter = JivagoBannerFilter()
        self.filter_chain_mock: FilterChain = mock.create_autospec(FilterChain)

    def test_whenApplyingFilter_thenAddsAHeaderField(self):
        request = RequestBuilder().build()
        response = ResponseBuilder().build()

        self.filter.doFilter(request, response, self.filter_chain_mock)

        self.assertTrue("Jivago" in response.headers['X-POWERED-BY'])



import os
import unittest

import test_data
from jivago.wsgi.request.partial_content_handler import PartialContentHandler
from test_utils.request_builder import RequestBuilder

TEST_FILE = os.path.join(os.path.dirname(test_data.__file__), "binary_file")
TEST_FILE_TOTAL_SIZE = os.path.getsize(TEST_FILE)


class PartialContentHandlerTest(unittest.TestCase):

    def setUp(self):
        self.partialContentHandler = PartialContentHandler()

    def test_givenHttpPartialRange_whenHandlingContentRequest_thenResponseHas206HttpStatus(self):
        partial_request = RequestBuilder().headers({"Range": "bytes=0-"}).build()

        response = self.partialContentHandler.handle_partial_content_request(partial_request, TEST_FILE)

        self.assertEqual(206, response.status)

    def test_givenStartingByteRange_whenHandlingContentRequest_thenReturnBytesStartingAtOffset(self):
        a_starting_offset = 6
        partial_request = RequestBuilder().headers({"Range": f"bytes={a_starting_offset}-"}).build()

        response = self.partialContentHandler.handle_partial_content_request(partial_request, TEST_FILE, max_block_size=1000000000)

        self.assertEqual(TEST_FILE_TOTAL_SIZE - a_starting_offset, len(response.body))
        with open(TEST_FILE, 'rb') as f:
            f.seek(6)
            self.assertEqual(f.read(), response.body)

    def test_givenMissingHttpRange_whenHandlingContentRequest_thenReturn200OkStatus(self):
        request = RequestBuilder().build()

        response = self.partialContentHandler.handle_partial_content_request(request, TEST_FILE)

        self.assertEqual(200, response.status)

    def test_givenMissingHttpRange_whenHandlingContentRequest_thenReturnTheWholeFile(self):
        request = RequestBuilder().build()

        response = self.partialContentHandler.handle_partial_content_request(request, TEST_FILE)

        with open(TEST_FILE, 'rb') as f:
            self.assertEqual(f.read(), response.body)

# TODO actual test with defined end offset



from jivago.lang.annotations import Override
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.invocation.url_encoded_form_parser import parse_urlencoded_form
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class HttpFormDeserializationFilter(Filter):

    @Override
    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        if request.headers['Content-Type'] == 'application/x-www-form-urlencoded':
            request.body = parse_urlencoded_form(request.body.decode("utf-8"))

        chain.doFilter(request, response)



import json
from typing import Type, TypeVar

from jivago.lang.annotations import Override, Serializable
from jivago.lang.registry import Registry
from jivago.serialization.deserializer import Deserializer
from jivago.serialization.serializer import Serializer

BUILTIN_TYPES = (str, float, int)

T = TypeVar('T')


class ObjectMapper(object):

    def __init__(self):
        self.deserializer = Deserializer(ObjectMapper.__RegistryStub())
        self.serializer = Serializer()

    def deserialize(self, json_str: str, clazz: Type[T]) -> T:
        dictionary = json.loads(json_str)
        return self.deserializer.deserialize(dictionary, clazz)

    def serialize(self, serializable: object) -> str:
        dictionary = self.serializer.serialize(serializable)
        return json.dumps(dictionary)

    class __RegistryStub(Registry):

        @Override
        def is_annotated(self, object: object, annotation: "Annotation"):
            return annotation == Serializable and object not in BUILTIN_TYPES + (dict,)



from typing import Dict, Type

from jivago.inject import typing_meta_helper
from jivago.lang.annotations import Override
from jivago.lang.stream import Stream
from jivago.serialization.deserialization_strategy import DeserializationStrategy, T


class TypedDictionaryDeserializationStrategy(DeserializationStrategy):

    def __init__(self, deserializer: "Deserializer"):
        self.deserializer = deserializer

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return typing_meta_helper.is_typing_meta_collection(declared_type, ('Dict',))

    @Override
    def deserialize(self, obj, declared_type: Type[Dict[str, T]]) -> dict:
        value_type = declared_type.__args__[1]
        return Stream(obj.items()) \
            .map(lambda k, v: (k, self.deserializer.deserialize(v, value_type))) \
            .toDict()



from typing import Callable, Optional

from jivago.inject import typing_meta_helper
from jivago.inject.exception.instantiation_exception import InstantiationException
from jivago.inject.exception.non_injectable_constructor_exception import NonInjectableConstructorException
from jivago.inject.scope_cache import ScopeCache
from jivago.lang.annotations import Inject
from jivago.lang.registry import Registry
from jivago.lang.stream import Stream


class ServiceLocator(object):
    def __init__(self, registry=Registry()):
        self.registry = registry
        self.literals = {}
        self.components = {}
        self.providers = {}
        self.scopeCaches = []

    def bind(self, interface, implementation):
        if self.__is_provider_function(implementation):
            self.providers[interface] = implementation
        elif isinstance(implementation, type):
            self.components[interface] = implementation
        else:
            self.literals[interface] = implementation

    def register_scope(self, scope_cache: ScopeCache):
        self.scopeCaches.append(scope_cache)

    def get(self, interface: type):
        if interface in self.literals.keys():
            return self.literals[interface]
        if typing_meta_helper.is_typing_meta_collection(interface):
            return self.get_all(interface.__args__[0])

        stored_component = Stream.of(self.components, self.providers) \
            .firstMatch(lambda x: interface in x) \
            .map(lambda x: x[interface]) \
            .orElseThrow(InstantiationException("Could not instantiate {}.".format(interface)))

        scope = self.__get_scope(stored_component)
        if scope is not None and scope.is_stored(stored_component):
            return scope.get(stored_component)
        instance = None
        
        if interface in self.providers.keys():
            instance = self.__inject_function(self.providers[interface])
        else:
            constructor = stored_component.__init__
            instance = self.__inject_constructor(stored_component, constructor)

        if scope:
            scope.store(stored_component, instance)
        return instance

    def get_all(self, clazz: type):
        return Stream(self.literals.keys(), self.components.keys(), self.providers.keys()).filter(
            lambda k: issubclass(k, clazz)).map(lambda k: self.get(k)).toList()

    def __is_provider_function(self, obj) -> bool:
        return callable(obj) and not isinstance(obj, type)

    def __inject_function(self, provider_method):
        parameters = []
        try:
            parameter_types = provider_method.__annotations__
            parameter_names = provider_method.__code__.co_varnames

            for name in parameter_names:
                if name not in parameter_types:
                    # This is a variable, for instance a function definition, and not an actual parameter
                    # Member variables are ordered after the function parameters, so when we get here, all parameters
                    # have been set already.
                    break
                parameter = parameter_types[name]
                parameters.append(self.get(parameter))
        except AttributeError:
            pass
        return provider_method(*parameters)

    def __inject_constructor(self, stored_component, constructor):
        if not self.__is_injectable(constructor):
            raise NonInjectableConstructorException(stored_component)

        the_object = object.__new__(stored_component)
        parameters = []
        try:
            parameter_types = constructor.__annotations__
            parameter_names = constructor.__code__.co_varnames[1::]

            for name in parameter_names:
                if name not in parameter_types:
                    # This is a variable, for instance a function definition, and not an actual parameter
                    # Member variables are ordered after the function parameters, so when we get here, all parameters
                    # have been set already.
                    break
                parameter = parameter_types[name]
                parameters.append(self.get(parameter))
        except AttributeError:
            pass
        constructor(the_object, *parameters)
        return the_object

    def __is_injectable(self, constructor_function: Callable) -> bool:
        try:
            return self.registry.is_annotated(constructor_function, Inject) or (
                    len(constructor_function.__code__.co_varnames) == 1 and
                    constructor_function.__code_co_varnames[0] == 'self')
        except AttributeError:
            # For object constructor wrapper, on classes which do not explicitly define a constructor
            return True

    def __get_scope(self, component: type) -> Optional[ScopeCache]:
        found_scope = Stream(self.scopeCaches).firstMatch(lambda scope: scope.handles_component(component))
        return found_scope.get() if found_scope.isPresent() else None



import unittest
from unittest import mock

from jivago.serialization.deserializer import Deserializer
from jivago.serialization.serializer import Serializer
from jivago.templating.rendered_view import RenderedView
from jivago.templating.template_filter import TemplateFilter
from jivago.templating.view_template_repository import ViewTemplateRepository
from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response

A_RESPONSE = Response.empty()

A_REQUEST = Request('GET', "/", {}, "", "")
A_TEMPLATE = "{{ foo }}"


class TemplateFilterTest(unittest.TestCase):

    def setUp(self):
        self.viewTemplateRepositoryMock: ViewTemplateRepository = mock.create_autospec(ViewTemplateRepository)
        self.serializerMock: Serializer = mock.create_autospec(Serializer)
        self.filterChainMock: FilterChain = mock.create_autospec(FilterChain)
        self.deserializerMock: Deserializer = mock.create_autospec(Deserializer)
        self.templateFilter = TemplateFilter(self.viewTemplateRepositoryMock, self.serializerMock, self.deserializerMock)

    def test_givenAResponseWhichShouldNotBeTemplated_whenApplyingFilter_thenDoNothing(self):
        returned_response = Response.empty()

        self.templateFilter.doFilter(A_REQUEST, returned_response, self.filterChainMock)

        self.assertEqual(A_RESPONSE.headers.items(), returned_response.headers.items())
        self.assertEqual(A_RESPONSE.body, returned_response.body)
        self.assertEqual(A_RESPONSE.status, returned_response.status)

    def test_givenATemplatedResponse_whenApplyingFilter_thenContentTypeIsSetToTextHtml(self):
        a_templated_response = Response(200, {}, RenderedView("test.html", {"foo": "bar"}))
        self.viewTemplateRepositoryMock.get_template.return_value = A_TEMPLATE

        self.templateFilter.doFilter(A_REQUEST, a_templated_response, self.filterChainMock)

        self.assertEqual("text/html", a_templated_response.headers['Content-Type'])

    def test_givenATemplatedResponse_whenApplyingFilter_thenTheTemplateIsProcessedByJinja2(self):
        a_templated_response = Response(200, {}, RenderedView("test.html", {"foo": "bar"}))
        self.deserializerMock.is_deserializable_type.return_value = False
        self.viewTemplateRepositoryMock.get_template.return_value = A_TEMPLATE

        self.templateFilter.doFilter(A_REQUEST, a_templated_response, self.filterChainMock)

        self.assertEqual("bar", a_templated_response.body)

    def test_givenADto_whenApplyingFilter_thenDtoIsSerializedToDictionaryBeforeRenderingTheTemplate(self):
        A_DTO = object()
        a_templated_response = Response(200, {}, RenderedView("test.html", A_DTO))
        self.serializerMock.serialize.return_value = {"foo": "bar"}
        self.viewTemplateRepositoryMock.get_template.return_value = A_TEMPLATE

        self.templateFilter.doFilter(A_REQUEST, a_templated_response, self.filterChainMock)

        self.assertEqual("bar", a_templated_response.body)



import os

from jivago.templating.no_such_template_exception import NoSuchTemplateException


class ViewTemplateRepository(object):

    def __init__(self, view_template_folder: str):
        self.view_template_folder = view_template_folder

    def get_template(self, filename: str) -> str:
        try:
            with open(os.path.join(self.view_template_folder, filename), 'r') as f:
                return "\n".join(f.readlines())
        except FileNotFoundError:
            raise NoSuchTemplateException()



from typing import Optional

from jivago.inject import typing_meta_helper
from jivago.lang.annotations import Override
from jivago.lang.stream import Stream
from jivago.serialization.deserialization_strategy import DeserializationStrategy, T


class OptionalAttributeDeserializationStrategy(DeserializationStrategy):

    def __init__(self, deserializer: "Deserializer"):
        self.deserializer = deserializer

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return typing_meta_helper.is_optional_typing_meta(declared_type)

    @Override
    def deserialize(self, obj, declared_type: Optional[T]) -> Optional[T]:
        if obj is None:
            return obj

        non_nil_declared_type = Stream(declared_type.__args__).firstMatch(lambda x: x != type(None))
        return self.deserializer.deserialize(obj, non_nil_declared_type)



from jivago.wsgi.request.partial_content_handler import PartialContentHandler
from jivago.wsgi.request.response import Response


class StaticFileServingResource(object):

    def __init__(self, filepath: str):
        self.filepath = filepath
        self.partial_content_handler = PartialContentHandler()

    def serve_file(self) -> Response:
        with open(self.filepath, 'rb') as f:
            return Response(200, {}, f.read())



from jivago.wsgi.request.request import Request


class ModifiedPathFlyweightRequest(Request):

    def __init__(self, request: Request, path: str):
        self.path = path
        self._request = request

    def __getattr__(self, item):
        return self._request.__getattribute__(item)



import unittest
from typing import List

from jivago.config.router.cors_rule import CorsRule
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.registry import Registry
from jivago.serialization.deserializer import Deserializer
from jivago.wsgi.invocation.rewrite.path_rewriting_route_handler_decorator import PathRewritingRouteHandlerDecorator
from jivago.wsgi.invocation.route_handler_factory import RouteHandlerFactory
from jivago.wsgi.methods import GET
from jivago.wsgi.routing.cors.cors_preflight_request_handler import CorsPreflightRequestHandler
from jivago.wsgi.routing.cors.cors_request_handler_factory import CorsRequestHandlerFactory
from jivago.wsgi.routing.exception.method_not_allowed_exception import MethodNotAllowedException
from jivago.wsgi.routing.exception.unknown_path_exception import UnknownPathException
from jivago.wsgi.routing.routing_rule import RoutingRule
from jivago.wsgi.routing.table.tree_routing_table import TreeRoutingTable
from test_utils.request_builder import RequestBuilder

RESOURCE_CLASS = object()

ROUTE_METHOD = object()

PATH = '/path'


class RouteHandlerFactoryTest(unittest.TestCase):

    def setUp(self):
        self.routing_table = TreeRoutingTable()
        self.routing_table.register_route(GET, PATH, RESOURCE_CLASS, ROUTE_METHOD)
        self.route_handler_factory = RouteHandlerFactory(ServiceLocator(), Deserializer(Registry()),
                                                         [RoutingRule(
                                                             "/", self.routing_table
                                                         )], CorsRequestHandlerFactory([CorsRule("/", {})]))

        self.request_builder = RequestBuilder().method("GET").path(PATH)

    def test_givenNoMatchingRoutes_whenCreatingRouteHandlers_thenRaiseUnknownPathException(self):
        request = self.request_builder.path("/unknown").build()

        with self.assertRaises(UnknownPathException):
            self.route_handler_factory.create_route_handlers(request)

    def test_givenDisallowedMethod_whenCreatingRouteHandlers_thenRaiseMethodNotAllowedException(self):
        request = self.request_builder.method("POST").build()

        with self.assertRaises(MethodNotAllowedException):
            self.route_handler_factory.create_route_handlers(request)

    def test_givenPrefixedRule_whenCreatingRouteHandlers_thenCreateResourceInvokerWithTruncatedPathForEveryRoute(self):
        request = self.request_builder.path("/prefix" + PATH).build()
        rule = RoutingRule("/prefix", self.routing_table)
        self.route_handler_factory = RouteHandlerFactory(ServiceLocator(), Deserializer(Registry()),
                                                         [rule], CorsRequestHandlerFactory([CorsRule("/", {})]))

        handlers: List[PathRewritingRouteHandlerDecorator] = [x for x in
                                                              self.route_handler_factory.create_route_handlers(request)]

        self.assertEqual(1, len(handlers))
        # TODO do assertion which does not imply internal structure
        self.assertEqual(PATH, handlers[0].route_handler.new_path)

    def test_givenCorsRequestOnUnknownPath_whenCreatingRouteHandlers_thenRaiseUnknownPathException(self):
        request = self.request_builder.method("OPTIONS").path("/unknown").build()

        with self.assertRaises(UnknownPathException):
            self.route_handler_factory.create_route_handlers(request)

    def test_givenCorsRequest_whenCreatingRouteHandlers_thenReturnSingleCorsHandler(self):
        request = self.request_builder.method("OPTIONS").build()

        handlers = [x for x in self.route_handler_factory.create_route_handlers(request)]

        self.assertEqual(1, len(handlers))
        self.assertIsInstance(handlers[0], CorsPreflightRequestHandler)



import unittest

from jivago.wsgi.invocation.url_encoded_form_parser import parse_urlencoded_form

A_METHOD = "GET"
A_PATH = "/hello"
HEADERS = {}
QUERY_STRING = "query=foobar&query2=barbaz"
BODY = ""


class UrlEncodedQueryParserTest(unittest.TestCase):

    def test_whenParsingQueryParameters_thenReturnQueryKeyPairsInADictionary(self):
        parameters = parse_urlencoded_form(QUERY_STRING)

        self.assertEqual(2, len(parameters))
        self.assertEqual("foobar", parameters['query'])
        self.assertEqual("barbaz", parameters['query2'])

    def test_givenRequestWithoutQueryParameters_whenParsingQueryParameters_thenReturnEmptyDictionary(self):
        parameters = parse_urlencoded_form("")

        self.assertEqual(0, len(parameters))



import unittest
from datetime import datetime
from typing import Optional, List, Dict, Tuple, Iterable, NamedTuple

from jivago.lang.annotations import Serializable
from jivago.lang.registry import Registry
from jivago.serialization.deserializer import Deserializer
from jivago.serialization.serialization_exception import SerializationException
from jivago.serialization.serializer import Serializer
from jivago.wsgi.invocation.incorrect_attribute_type_exception import IncorrectAttributeTypeException

OBJECT_WITH_UNKNOWN_PROPERTY = {"name": "a_name", "unknown-property": "foobar"}

OBJECT_WITH_MISSING_VALUES = {}


class DtoSerializationHandler(object):
    def __init__(self, registry: Registry):
        self.deserializer = Deserializer(registry)
        self.serializer = Serializer()

    def deserialize(self, object, type):
        return self.deserializer.deserialize(object, type)

    def serialize(self, obj):
        return self.serializer.serialize(obj)


class DtoSerializationHandlerTest(unittest.TestCase):

    def setUp(self):
        self.serialization_handler = DtoSerializationHandler(Registry())

    def test_whenDeserializing_thenCreateDtoMatchingDictionaryItems(self):
        a_dto = self.serialization_handler.deserialize({"name": "a name"}, ADto)

        self.assertIsInstance(a_dto, ADto)
        self.assertEqual("a name", a_dto.name)

    def test_givenDtoWithDefinedConstructor_whenDeserializing_thenInstantiateByInvokingInitializer(self):
        a_dto = self.serialization_handler.deserialize({"name": "a name"}, ADtoWithAConstructor)

        self.assertIsInstance(a_dto, ADtoWithAConstructor)
        self.assertEqual("a name", a_dto.name)

    def test_givenUnknownProperty_whenDeserializing_thenIgnoreTheUnknownProperty(self):
        dto = self.serialization_handler.deserialize(OBJECT_WITH_UNKNOWN_PROPERTY, ADto)

        self.assertIsInstance(dto, ADto)
        self.assertFalse("unknown-property" in dto.__dict__)

    def test_givenMissingProperty_whenDeserializing_thenRaiseIncorrectAttributeException(self):
        with self.assertRaises(IncorrectAttributeTypeException):
            self.serialization_handler.deserialize(OBJECT_WITH_MISSING_VALUES, ADto)

    def test_givenMissingKey_whenDeserializingOptionalParameter_thenAssignNoneToTheAttribute(self):
        dto = self.serialization_handler.deserialize(OBJECT_WITH_MISSING_VALUES, ADtoWithOptionalValue)

        self.assertIsInstance(dto, ADtoWithOptionalValue)
        self.assertIsNone(dto.name)

    def test_givenNestedDtos_whenSerializing_thenRecursivelySerializeDtos(self):
        child = ChildDto()
        child.name = "a name"
        nested = A_NESTED_DTO
        nested.child_dto = child

        dictionary = self.serialization_handler.serialize(nested)

        self.assertEqual({"child_dto": {"name": "a name"}}, dictionary)

    def test_givenNestedDtoCollection_whenSerializing_thenRecursivelySerializeDtos(self):
        children = [ChildDto() for x in range(0, 2)]
        for child in children:
            child.name = "a name"
        collection = ACollectionDto(children)

        dictionary = self.serialization_handler.serialize(collection)
        self.assertEqual({"children": [{"name": "a name"}, {"name": "a name"}]}, dictionary)

    def test_givenNestedDictionary_whenDeserializing_thenRecursivelyDeserializeToDto(self):
        dictionary = {"children": [{"name": "a name"}, {"name": "a name"}]}

        dto = self.serialization_handler.deserialize(dictionary, ACollectionDto)

        self.assertIsInstance(dto, ACollectionDto)
        self.assertEqual("a name", dto.children[1].name)

    def test_givenNonSerializableObject_whenSerializing_thenThrowSerializationException(self):
        non_serializable_object = object()

        with self.assertRaises(SerializationException):
            self.serialization_handler.serialize(non_serializable_object)

    def test_givenDictionaryOfDtos_whenSerializing_thenSerializeValuesIntoDictionaries(self):
        dictionary_of_dtos = {"foobar": A_NESTED_DTO}

        dictionary = self.serialization_handler.serialize(dictionary_of_dtos)

        self.assertEqual({"foobar": {"child_dto": {"name": "a name"}}}, dictionary)

    def test_givenListTypingMeta_whenDeserializing_thenReturnListOfDtos(self):
        serialized_dtos = [{"name": "foobar"}]

        dtos = self.serialization_handler.deserialize(serialized_dtos, List[ADto])

        self.assertEqual(1, len(dtos))
        self.assertEqual("foobar", dtos[0].name)

    def test_givenTwiceNestedDto_whenDeserializing_thenDeserializeDto(self):
        dto: TwiceNestedDto = self.serialization_handler.deserialize({"nested_dto": {"child_dto": {"name": "foobar"}}},
                                                                     TwiceNestedDto)

        self.assertEqual("foobar", dto.nested_dto.child_dto.name)

    def test_givenBodyAlreadyInTheRightType_whenDeserializing_thenReturnObjectAsIs(self):
        result = self.serialization_handler.deserialize(5, int)

        self.assertEqual(5, result)

    def test_givenIncorrectParameterTypes_whenInjectingConstructor_thenRaiseIncorrectAttributeTypeException(self):
        with self.assertRaises(IncorrectAttributeTypeException):
            self.serialization_handler.deserialize({"children": 5}, ACollectionDto)

    def test_givenDeclaredTypedDictionary_whenDeserializing_thenCreateProperlyPopulatedDictionary(self):
        result: NestedTypeDictDto = self.serialization_handler.deserialize(
            {"children": {"first": {"name": "foo"}, "second": {"name": "bar"}}}, NestedTypeDictDto)

        self.assertIsInstance(result.children, dict)
        self.assertEqual("foo", result.children["first"].name)
        self.assertEqual("bar", result.children["second"].name)

    def test_givenTypedDictTargetClass_whenDeserializing_thenCreateProperlyPopulatedDictionary(self):
        result = self.serialization_handler.deserialize({"1": 1}, Dict[str, int])

        self.assertIsInstance(result, dict)
        self.assertEqual(1, result["1"])

    def test_givenDtoWithIterables_whenDeserializing_thenPopulateIterableFieldWithList(self):
        result: DtoWithIterablesAndTuples = self.serialization_handler.deserialize(
            {"tuples": ["1", "2"], "iterables": ["1", "2"]}, DtoWithIterablesAndTuples)

        self.assertIsInstance(result.iterables, list)

    def test_givenDtoWithTuples_whenDeserializing_thenPopulateFieldWithTuple(self):
        result: DtoWithIterablesAndTuples = self.serialization_handler.deserialize(
            {"tuples": ["1", "2"], "iterables": ["1", "2"]}, DtoWithIterablesAndTuples)

        self.assertIsInstance(result.tuples, tuple)

    def test_givenTuple_whenSerializing_thenSerializeToListInJson(self):
        tuple_response = ("foo",)

        serialized = self.serialization_handler.serialize(tuple_response)

        self.assertEqual(['foo'], serialized)

    def test_givenDatetime_whenSerializing_thenSerializeToStringInJson(self):
        date = datetime(year=1984, month=1, day=24)
        result = self.serialization_handler.serialize({"date": date})

        self.assertEqual(result["date"], str(date))

    def test_givenDatetime_whenDeserializing_thenDeserializeFromString(self):
        serialized = "1984-01-24"

        result = self.serialization_handler.deserialize(serialized, datetime)

        self.assertIsInstance(result, datetime)
        self.assertEqual(datetime(1984, 1, 24), result)

    def test_givenNamedTuple_whenDeserializing_thenDeserializeFromDictionary(self):
        serialized = {"name": "paul atreides", "age": 17}

        result = self.serialization_handler.deserialize(serialized, ANamedTuple)

        self.assertIsInstance(result, ANamedTuple)

    def test_givenNoneAttribute_whenSerializing_thenSerializesToNoneOrNull(self):
        expected = {"name": None}

        result = self.serialization_handler.serialize(expected)

        self.assertEqual(None, result["name"])

    def test_whenSerializing_DoesNotModifyTheSourceObject(self):
        given = ACollectionDto([ChildDto()])

        self.serialization_handler.serialize(given)

        self.assertIsInstance(given.children[0], ChildDto)

    def test_givenATypeDerivedFromABuiltinType_whenSerializing_thenSerializesAsBuiltinType(self):
        given = {"name": DerivedString("my-derived-name")}

        result = self.serialization_handler.serialize(given)

        self.assertEqual("my-derived-name", result["name"])

    def test_givenATypeDerivedFromABuiltingType_whenDeserializing_thenDeserializesAsBuiltinType(self):
        given = {"name": "my-name"}

        result: DtoWithDerivedStringMember = self.serialization_handler.deserialize(given, DtoWithDerivedStringMember)

        self.assertEqual("my-name", result.name)
        self.assertIsInstance(result.name, DerivedString)


@Serializable
class ADto(object):
    name: str


@Serializable
class ADtoWithAConstructor(object):
    def __init__(self, name: str):
        self.name = name


@Serializable
class ADtoWithOptionalValue(object):
    name: Optional[str]


@Serializable
class ChildDto(object):
    name: str


@Serializable
class ANestedDto(object):
    child_dto: ChildDto


@Serializable
class ACollectionDto(object):
    children: List[ChildDto]

    def __init__(self, children: List[ChildDto]):
        self.children = children


@Serializable
class TwiceNestedDto(object):
    nested_dto: ANestedDto


@Serializable
class NestedTypeDictDto(object):
    children: Dict[str, ChildDto]


@Serializable
class DtoWithIterablesAndTuples(object):
    tuples: Tuple[str]
    iterables: Iterable[str]


class ANamedTuple(NamedTuple):
    name: str
    age: int


class DerivedString(str):
    pass


@Serializable
class DtoWithDerivedStringMember(object):
    name: DerivedString


A_NESTED_DTO = ANestedDto()
A_NESTED_DTO.child_dto = ChildDto()
A_NESTED_DTO.child_dto.name = "a name"



from typing import TypeVar

T = TypeVar("T")


class SerializationStrategy(object):

    def can_handle_serialization(self, obj) -> bool:
        raise NotImplementedError

    def serialize(self, obj) -> object:
        raise NotImplementedError



import unittest

import os

import test_data
from jivago.templating.no_such_template_exception import NoSuchTemplateException
from jivago.templating.view_template_repository import ViewTemplateRepository

TEMPLATE_CONTENT = "{{ data }}\n"


class ViewTemplateRepositoryTest(unittest.TestCase):

    def setUp(self):
        self.templateRepository = ViewTemplateRepository(os.path.dirname(test_data.__file__))

    def test_whenGettingTemplate_thenReturnCorrespondingFileContent(self):
        template = self.templateRepository.get_template("template.txt")

        self.assertEqual(TEMPLATE_CONTENT, template)

    def test_givenInexistentTemplateFile_whenGettingTemplate_thenThrowNoSuchTemplateException(self):
        with self.assertRaises(NoSuchTemplateException):
            self.templateRepository.get_template("inexistent-file.html")



import unittest

from jivago.wsgi.request.headers import Headers
from jivago.wsgi.routing.cors.cors_preflight_request_handler import CorsPreflightRequestHandler
from test_utils.request_builder import RequestBuilder


class CorsPreflightRequestHandlerTest(unittest.TestCase):

    def setUp(self):
        self.cors_headers = Headers({"Access-Control-Allow-Origin": "*"})
        self.cors_handler = CorsPreflightRequestHandler(self.cors_headers)

    def test_givenAllowEverythingOriginHeader_whenHandlingPreflight_thenReturn200OK(self):
        request = RequestBuilder().headers({'Origin': 'some-origin'}).build()

        response = self.cors_handler.invoke(request)

        self.assertEqual(200, response.status)

    def test_givenOriginThatStartsWithAllowedHeader_whenHandlingPreflight_thenReturn200OKWithVaryHeader(self):
        self.cors_headers['Access-Control-Allow-Origin'] = "http://api.jivago.io"
        request = RequestBuilder().headers({'Origin': 'http://api.jivago.io'}).build()
        self.cors_handler = CorsPreflightRequestHandler(self.cors_headers)

        response = self.cors_handler.invoke(request)

        self.assertEqual(200, response.status)
        self.assertEqual('Origin', response.headers['Vary'])

    def test_givenNonMatchingOrigin_whenHandlingPreflight_thenReturn400BadRequest(self):
        request = RequestBuilder().headers({'Origin': 'http://foobar.com'}).build()
        self.cors_handler = CorsPreflightRequestHandler(Headers({'Access-Control-Allow-Origin': 'http://api.jivago.io'}))

        response = self.cors_handler.invoke(request)

        self.assertEqual(400, response.status)



from jivago.lang.annotations import Override

from jivago.serialization.deserializer import Deserializer
from jivago.serialization.serialization_exception import SerializationException
from jivago.wsgi.invocation.missing_route_invocation_argument import MissingRouteInvocationArgument
from jivago.wsgi.invocation.parameter_selection.parameter_selector import ParameterSelector
from jivago.wsgi.request.request import Request


class SerializedParameterSelector(ParameterSelector):

    def __init__(self, deserializer: Deserializer):
        self.deserializer = deserializer

    @Override
    def matches(self, paramater_declaration: type) -> bool:
        return self.deserializer.is_deserializable_type(paramater_declaration)

    @Override
    def format_parameter(self, parameter_name: str, parameter_declaration: type, request: Request) -> object:
        try:
            return self.deserializer.deserialize(request.body, parameter_declaration)
        except SerializationException:
            raise MissingRouteInvocationArgument()



import unittest

from jivago.wsgi.invocation.resource_invoker import ResourceInvoker
from jivago.wsgi.invocation.rewrite.path_rewriting_route_handler_decorator import PathRewritingRouteHandlerDecorator
from jivago.wsgi.methods import GET
from jivago.wsgi.routing.routing_rule import RoutingRule
from jivago.wsgi.routing.table.tree_routing_table import TreeRoutingTable
from test_utils.request_builder import RequestBuilder


class RoutingRuleTest(unittest.TestCase):

    def setUp(self):
        self.routing_table = TreeRoutingTable()
        self.routing_table.register_route(GET, "/path", object(), object())

    def test_matchingPrefixRootPrefix(self):
        rule = RoutingRule("/", None)

        self.assertTrue(rule.matches("/foo"))
        self.assertTrue(rule.matches("/"))

    def test_matchingPrefix(self):
        rule = RoutingRule("/users", None)

        self.assertTrue(rule.matches("/users"))
        self.assertTrue(rule.matches("/users/foo/bar/delete"))
        self.assertFalse(rule.matches("/admin"))

    def test_givenPathWhichDoesNotMatchRule_whenGettingRouteRegistrations_thenReturnEmptyList(self):
        rule = RoutingRule("/users", None)

        routes = rule.get_route_registrations("/foobar")

        self.assertEqual(0, len(routes))

    def test_whenCreatingRouteHandlers_thenRewritePaths(self):
        request = RequestBuilder().path("/users/path").build()
        rule = RoutingRule("/users", self.routing_table, rewrite_path=True)

        handlers = [x for x in rule.create_route_handlers(request, None, None)]

        self.assertIsInstance(handlers[0], PathRewritingRouteHandlerDecorator)

    def test_givenShouldNotRewritePath_whenCreatingRouteHandlers_thenReturnStandardResourceInvokers(self):
        request = RequestBuilder().path("/users/path").build()
        rule = RoutingRule("/users", self.routing_table, rewrite_path=False)

        handlers = [x for x in rule.create_route_handlers(request, None, None)]

        self.assertIsInstance(handlers[0], ResourceInvoker)



from jivago.config.abstract_binder import AbstractBinder
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Override
from jivago.wsgi.routing.cors.no_matching_cors_rule_exception_mapper import NoMatchingCorsRuleExceptionMapper
from jivago.wsgi.routing.exception.incorrect_resource_parameters_exception_mapper import \
    IncorrectResourceParametersExceptionMapper
from jivago.wsgi.routing.exception.method_not_allowed_exception_mapper import MethodNotAllowedExceptionMapper
from jivago.wsgi.routing.exception.unknown_path_exception_mapper import UnknownPathExceptionMapper


class ExceptionMapperBinder(AbstractBinder):

    @Override
    def bind(self, service_locator: ServiceLocator):
        service_locator.bind(MethodNotAllowedExceptionMapper, MethodNotAllowedExceptionMapper)
        service_locator.bind(UnknownPathExceptionMapper, UnknownPathExceptionMapper)
        service_locator.bind(IncorrectResourceParametersExceptionMapper, IncorrectResourceParametersExceptionMapper)
        service_locator.bind(NoMatchingCorsRuleExceptionMapper, NoMatchingCorsRuleExceptionMapper)



import json
import unittest
from unittest import mock

from jivago.wsgi.filter.filter_chain import FilterChain
from jivago.wsgi.request.json_serialization_filter import JsonSerializationFilter
from test_utils.request_builder import RequestBuilder
from test_utils.response_builder import ResponseBuilder

A_REQUEST = RequestBuilder().build()


class JsonSerializationFilterTest(unittest.TestCase):

    def setUp(self):
        self.filter = JsonSerializationFilter()
        self.filterChainMock: FilterChain = mock.create_autospec(FilterChain)

    def test_givenADictionary_whenApplyingFilter_thenSerializeToJson(self):
        body = {"hello": "goodbye"}
        response = ResponseBuilder().body(body).build()

        self.filter.doFilter(A_REQUEST, response, self.filterChainMock)

        self.assertEqual(body, json.loads(response.body))

    def test_givenAList_whenApplyingFilter_thenSerializeToJson(self):
        body = ["hello"]
        response = ResponseBuilder().body(body).build()

        self.filter.doFilter(A_REQUEST, response, self.filterChainMock)

        self.assertEqual(body, json.loads(response.body))

    def test_givenSerializable_whenApplyingFilter_thenSetContentTypeToApplicationJson(self):
        body = {"hello": "goodbye"}
        response = ResponseBuilder().body(body).build()

        self.filter.doFilter(A_REQUEST, response, self.filterChainMock)

        self.assertEqual("application/json", response.headers['Content-Type'])

    def test_givenApplicationJsonHeaderWithEmptyBody_whenApplyingFilter_thenDoNotDeserializeTheEmptyString(self):
        request = RequestBuilder().headers({"Content-Type": "application/json"}).body("").method("POST").build()

        self.filter.doFilter(request, ResponseBuilder().build(), self.filterChainMock)

        self.assertEqual("", request.body)

    def test_givenMultipleContentTypeDirectives_whenApplyingFilter_thenMatchOnMediaType(self):
        """https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type"""
        request = RequestBuilder() \
            .headers({"Content-Type": "application/json; charset=utf-8"}) \
            .body('{"name": "bar"}') \
            .build()

        self.filter.doFilter(request, ResponseBuilder().build(), self.filterChainMock)

        self.assertEqual({"name": "bar"}, request.body)

    def test_givenTrailingSemiColonInContentType_whenApplyingFilter_thenMatchAnyway(self):
        request = RequestBuilder() \
            .headers({"Content-Type": "application/json; "}) \
            .body('{"name": "bar"}') \
            .build()

        self.filter.doFilter(request, ResponseBuilder().build(), self.filterChainMock)

        self.assertEqual({"name": "bar"}, request.body)



from jivago.lang.stream import Stream


def parse_urlencoded_form(query_string: str) -> dict:
    pairs = query_string.split("&")
    return Stream(pairs).filter(lambda x: x != "").map(lambda pair: pair.split("=")).toDict()



from datetime import datetime
from typing import Optional

from jivago.lang.registry import ParametrizedAnnotation


@ParametrizedAnnotation
def Scheduled(cron: Optional[str], every: Optional["Duration"], start: Optional[datetime]):
    return lambda x: x


class _Interval(object):

    def __init__(self, interval_time_in_seconds: int):
        self.interval_time_in_seconds = interval_time_in_seconds


class Duration(object):
    SECOND = _Interval(1)
    MINUTE = _Interval(60)
    HOUR = _Interval(3600)
    DAY = _Interval(86400)



from typing import Type

from jivago.lang.annotations import Override
from jivago.serialization.deserialization_strategy import DeserializationStrategy
from jivago.wsgi.invocation.incorrect_attribute_type_exception import IncorrectAttributeTypeException


class ListDeserializationStrategy(DeserializationStrategy):

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return declared_type == list

    @Override
    def deserialize(self, obj, declared_type: Type[list]) -> list:
        if isinstance(obj, list):
            return obj
        raise IncorrectAttributeTypeException()



from typing import List

from jivago.config.router.router_builder import RouterBuilder
from jivago.inject.service_locator import ServiceLocator


class AbstractContext(object):
    INSTANCE: "AbstractContext" = None

    def __init__(self):
        self.serviceLocator = ServiceLocator()
        self.INSTANCE = self
        self.serviceLocator.bind(ServiceLocator, self.serviceLocator)

    def configure_service_locator(self):
        raise NotImplementedError

    def service_locator(self) -> ServiceLocator:
        return self.serviceLocator

    def get_views_folder_path(self) -> str:
        raise NotImplementedError

    def get_config_file_locations(self) -> List[str]:
        raise NotImplementedError

    def create_router_config(self) -> RouterBuilder:
        raise NotImplementedError

    def get_banner(self) -> List[str]:
        return """       _ _                        
      | (_)                       
      | |___   ____ _  __ _  ___  
  _   | | \ \ / / _` |/ _` |/ _ \ 
 | |__| | |\ V / (_| | (_| | (_) |
  \____/|_| \_/ \__,_|\__, |\___/ 
                       __/ |      
                      |___/       """.split("\n")



import unittest
from datetime import datetime
from unittest import mock

from jivago.lang.registry import Registry
from jivago.lang.runnable import Runnable
from jivago.scheduling.annotations import Scheduled, Duration
from jivago.scheduling.cron_schedule import CronSchedule
from jivago.scheduling.task_schedule_initializer import TaskScheduleInitializer
from jivago.scheduling.task_scheduler import TaskScheduler

TIME_INTERVAL = Duration.MINUTE

DATE_IN_THE_FUTURE = datetime(2077, 5, 5)

CRON_STRING = "* * * * * *"


class TaskScheduleInitializerTest(unittest.TestCase):

    def setUp(self):
        self.registry = Registry()
        self.taskScheduleInitializer = TaskScheduleInitializer(self.registry, "")
        self.taskSchedulerMock: TaskScheduler = mock.create_autospec(TaskScheduler)

    def test_givenCronString_whenCreatingSchedule_thenScheduleWithCron(self):
        self.registry.content = {}
        self.registry.register(Scheduled, ScheduledClass, arguments={"cron": CRON_STRING})

        self.taskScheduleInitializer.initialize_task_scheduler(self.taskSchedulerMock)

        runnable_class, schedule = self.taskSchedulerMock.schedule_task.call_args[0]
        self.assertIsInstance(schedule, CronSchedule)

    def test_givenCronStartDate_whenCreatingSchedule_thenDoNotScheduleBeforeStartTime(self):
        self.registry.content = {}
        self.registry.register(Scheduled, ScheduledClass, arguments={"cron": CRON_STRING, "start": DATE_IN_THE_FUTURE})

        self.taskScheduleInitializer.initialize_task_scheduler(self.taskSchedulerMock)

        runnable_class, schedule = self.taskSchedulerMock.schedule_task.call_args[0]
        self.assertTrue(schedule.next_start_time() > DATE_IN_THE_FUTURE)

    def test_givenSimpleTimeInterval_whenCreatingSchedule_thenScheduleEveryXSeconds(self):
        self.registry.content = {}
        self.registry.register(Scheduled, ScheduledClass, arguments={"every": TIME_INTERVAL})

        self.taskScheduleInitializer.initialize_task_scheduler(self.taskSchedulerMock)

        runnable_class, schedule = self.taskSchedulerMock.schedule_task.call_args[0]
        self.assertEqual(TIME_INTERVAL.interval_time_in_seconds, schedule.interval)

    def test_givenSimpleTimeIntervalWithStartTime_whenCreatingSchedule_thenDoNotScheduleBeforeStartTime(self):
        self.registry.content = {}
        self.registry.register(Scheduled, ScheduledClass, arguments={"every": TIME_INTERVAL, "start": DATE_IN_THE_FUTURE})

        self.taskScheduleInitializer.initialize_task_scheduler(self.taskSchedulerMock)

        runnable_class, schedule = self.taskSchedulerMock.schedule_task.call_args[0]
        self.assertTrue(schedule.next_start_time() > DATE_IN_THE_FUTURE)


class ScheduledClass(Runnable):
    def run(self):
        pass



import os
from typing import List, Type, Union

from jivago.config.abstract_context import AbstractContext
from jivago.config.exception_mapper_binder import ExceptionMapperBinder
from jivago.config.router.filtering.annotation import RequestFilter
from jivago.config.router.filtering.auto_discovering_filtering_rule import AutoDiscoveringFilteringRule
from jivago.config.router.filtering.filtering_rule import FilteringRule
from jivago.config.router.router_builder import RouterBuilder
from jivago.config.startup_hooks import Init, PreInit, PostInit
from jivago.event.async_event_bus import AsyncEventBus
from jivago.event.config.annotations import EventHandlerClass
from jivago.event.config.reflective_event_bus_initializer import ReflectiveEventBusInitializer
from jivago.event.config.runnable_event_handler_binder import RunnableEventHandlerBinder
from jivago.event.event_bus import EventBus
from jivago.event.synchronous_event_bus import SynchronousEventBus
from jivago.inject.annotation import Component, Singleton
from jivago.inject.annoted_class_binder import AnnotatedClassBinder
from jivago.inject.provider_binder import ProviderBinder
from jivago.inject.scope_cache import ScopeCache
from jivago.lang.annotations import Override, BackgroundWorker
from jivago.lang.registry import Registry
from jivago.lang.stream import Stream
from jivago.scheduling.annotations import Scheduled
from jivago.scheduling.task_scheduler import TaskScheduler
from jivago.serialization.deserializer import Deserializer
from jivago.serialization.object_mapper import ObjectMapper
from jivago.serialization.serializer import Serializer
from jivago.templating.template_filter import TemplateFilter
from jivago.templating.view_template_repository import ViewTemplateRepository
from jivago.wsgi.annotations import Resource
from jivago.wsgi.filter.filter import Filter
from jivago.wsgi.filter.system_filters.body_serialization_filter import BodySerializationFilter
from jivago.wsgi.filter.system_filters.error_handling.application_exception_filter import ApplicationExceptionFilter
from jivago.wsgi.filter.system_filters.error_handling.unknown_exception_filter import UnknownExceptionFilter
from jivago.wsgi.filter.system_filters.jivago_banner_filter import JivagoBannerFilter
from jivago.wsgi.request.http_form_deserialization_filter import HttpFormDeserializationFilter
from jivago.wsgi.request.http_status_code_resolver import HttpStatusCodeResolver
from jivago.wsgi.request.json_serialization_filter import JsonSerializationFilter
from jivago.wsgi.request.partial_content_handler import PartialContentHandler
from jivago.wsgi.routing.routing_rule import RoutingRule
from jivago.wsgi.routing.table.auto_discovering_routing_table import AutoDiscoveringRoutingTable


class ProductionJivagoContext(AbstractContext):

    def __init__(self, root_package: "Module", registry: Registry, banner: bool = True):
        self.banner = banner
        self.root_package_name = root_package.__name__ if root_package else ''
        self.root_package = root_package
        self.registry = registry
        super().__init__()

    @Override
    def configure_service_locator(self):
        AnnotatedClassBinder(self.root_package_name, self.registry, Component).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, Resource).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, BackgroundWorker).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, Scheduled).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, Init).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, PreInit).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, PostInit).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, EventHandlerClass).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, RequestFilter).bind(self.serviceLocator)

        ProviderBinder(self.root_package_name, self.registry).bind(self.serviceLocator)
        for scope in self.scopes():
            scoped_classes = Stream(self.registry.get_annotated_in_package(scope, self.root_package_name)).map(
                lambda registration: registration.registered).toList()
            cache = ScopeCache(scope, scoped_classes)
            self.serviceLocator.register_scope(cache)

        Stream(self.get_default_filters()).forEach(lambda f: self.serviceLocator.bind(f, f))

        # TODO better way to handle Jivago Dependencies
        self.serviceLocator.bind(Registry, Registry.INSTANCE)
        self.serviceLocator.bind(TaskScheduler, TaskScheduler(self.serviceLocator))
        self.serviceLocator.bind(Deserializer, Deserializer(Registry.INSTANCE))
        self.serviceLocator.bind(Serializer, Serializer())
        self.serviceLocator.bind(ViewTemplateRepository, ViewTemplateRepository(self.get_views_folder_path()))
        self.serviceLocator.bind(BodySerializationFilter, BodySerializationFilter)
        self.serviceLocator.bind(PartialContentHandler, PartialContentHandler)
        self.serviceLocator.bind(HttpStatusCodeResolver, HttpStatusCodeResolver)
        self.serviceLocator.bind(ObjectMapper, ObjectMapper)
        self.serviceLocator.bind(EventBus, self.create_event_bus())
        self.serviceLocator.bind(SynchronousEventBus, self.serviceLocator.get(EventBus))
        self.serviceLocator.bind(AsyncEventBus, AsyncEventBus(self.serviceLocator.get(EventBus)))

        RunnableEventHandlerBinder(self.root_package_name, self.registry).bind(self.service_locator())

        ExceptionMapperBinder().bind(self.serviceLocator)

    def scopes(self) -> List[type]:
        return [Singleton, BackgroundWorker]

    @Override
    def get_views_folder_path(self) -> str:
        return os.path.join(os.path.dirname(self.root_package.__file__), "views") if self.root_package else ''

    @Override
    def get_config_file_locations(self) -> List[str]:
        return ["application.yml", "application.json", "properties.yml", "properties.json"]

    @Override
    def create_router_config(self) -> RouterBuilder:
        return RouterBuilder() \
            .add_rule(FilteringRule("*", self.get_default_filters())) \
            .add_rule(AutoDiscoveringFilteringRule("*", self.registry, self.root_package_name)) \
            .add_rule(RoutingRule("/", AutoDiscoveringRoutingTable(self.registry, self.root_package_name)))

    def get_default_filters(self) -> List[Union[Filter, Type[Filter]]]:
        default_filters = [UnknownExceptionFilter, TemplateFilter, JsonSerializationFilter,
                           HttpFormDeserializationFilter,
                           BodySerializationFilter, ApplicationExceptionFilter]
        if self.banner:
            default_filters.append(JivagoBannerFilter)

        return default_filters

    def create_event_bus(self) -> EventBus:
        return ReflectiveEventBusInitializer(self.service_locator(), self.registry,
                                             self.root_package_name).create_message_bus()



from jivago.lang.annotations import Override
from jivago.wsgi.filter.system_filters.error_handling.exception_mapper import ExceptionMapper
from jivago.wsgi.invocation.incorrect_resource_parameters_exception import IncorrectResourceParametersException
from jivago.wsgi.request.response import Response


class IncorrectResourceParametersExceptionMapper(ExceptionMapper):

    @Override
    def handles(self, exception: Exception) -> bool:
        return isinstance(exception, IncorrectResourceParametersException)

    @Override
    def create_response(self, exception: Exception) -> Response:
        return Response(400, {}, {"message": "Incorrect parameters"})



from typing import Type

from jivago.lang.annotations import Override
from jivago.lang.stream import Stream
from jivago.serialization.deserialization_strategy import DeserializationStrategy, T
from jivago.wsgi.invocation.incorrect_attribute_type_exception import IncorrectAttributeTypeException

BUILTIN_TYPES = (str, float, int, bool)


class BuiltInTypeDeserializationStrategy(DeserializationStrategy):

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return type(declared_type) == type and issubclass(declared_type, BUILTIN_TYPES)

    @Override
    def deserialize(self, obj, declared_type: Type[T]) -> T:
        if obj is None:
            raise IncorrectAttributeTypeException()
        return declared_type(obj)



from urllib.parse import unquote

from jivago.wsgi.invocation.url_encoded_form_parser import parse_urlencoded_form
from jivago.wsgi.methods import to_method
from jivago.wsgi.request.headers import Headers


class Request(object):
    def __init__(self, method: str, path: str, headers: Headers, query_string: str, body):
        self.method = method
        self.method_annotation = to_method(method)
        self.path = unquote(path)
        self.headers = headers
        self.body = body
        self.queryString = unquote(query_string)
        self._query_form = None

    @property
    def query_form(self) -> dict:
        if self._query_form is None:
            self._query_form = parse_urlencoded_form(self.queryString)
        return self._query_form



from typing import Type

from jivago.lang.annotations import Override
from jivago.lang.stream import Stream
from jivago.serialization.deserialization_strategy import DeserializationStrategy


class TupleDeserializationStrategy(DeserializationStrategy):

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return declared_type == tuple

    @Override
    def deserialize(self, obj, declared_type: Type[tuple]) -> tuple:
        return Stream(obj).toTuple()



def is_typing_meta_collection(clazz, metas=('List', 'Collection', 'Iterable', 'Tuple')) -> bool:
    if not hasattr(clazz, "__module__"):
        return False
    typing_meta_name = clazz._name if hasattr(clazz, "_name") else clazz.__name__ if hasattr(clazz, "__name__") else ""
    if typing_meta_name is None: # Python3.7 Union metas
        typing_meta_name = ""
    return clazz.__module__ == 'typing' and hasattr(clazz, '__args__') and typing_meta_name in metas


def is_union_typing_meta(clazz):
    return is_typing_meta_collection(clazz, ("",))

def is_optional_typing_meta(clazz):
    return is_union_typing_meta(clazz) and type(None) in clazz.__args__



from typing import Type

from jivago.lang.annotations import Override
from jivago.serialization.deserialization_strategy import DeserializationStrategy
from jivago.wsgi.invocation.incorrect_attribute_type_exception import IncorrectAttributeTypeException


class DictionaryDeserializationStrategy(DeserializationStrategy):

    @Override
    def can_handle_deserialization(self, declared_type: type) -> bool:
        return declared_type == dict

    @Override
    def deserialize(self, obj, declared_type: Type[dict]) -> dict:
        if obj is None:
            raise IncorrectAttributeTypeException()
        return obj



from typing import Callable

from jivago.lang.registry import Annotation


@Annotation
def Singleton(wrapped_class: type) -> type:
    return wrapped_class


@Annotation
def Component(wrapped_class: type) -> type:
    return wrapped_class


@Annotation
def Provider(wrapped_function: Callable) -> Callable:
    return wrapped_function



from jivago.lang.annotations import Override
from jivago.wsgi.invocation.parameter_selection.parameter_selector import ParameterSelector
from jivago.wsgi.request.request import Request


class RawRequestParameterSelector(ParameterSelector):

    @Override
    def matches(self, paramater_declaration: type) -> bool:
        return paramater_declaration == Request

    @Override
    def format_parameter(self, parameter_name: str, parameter_declaration: type, request: Request) -> object:
        return request



import inspect
import itertools
from typing import Iterable, Callable, Iterator, Tuple, Any, TypeVar, Generic, Set, List

from jivago.lang.nullable import Nullable

T = TypeVar('T')
S = TypeVar('S')


class Stream(Generic[T]):
    """Stream class to perform functional-style operations in an aesthetically-pleasing manner.

    Args:
        *iterables (Iterable) : Source iterables for the Stream object. When multiple iterables are given, they will be concatenated.
    """

    def __init__(self, *iterables: Iterable[T]):
        self.iterable = itertools.chain(*iterables)

    def map(self, fun: Callable[[T], S]) -> "Stream[S]":
        """Maps elements using the supplied function. When iterating over tuples, the function can take multiple arguments."""
        if self.__should_expand(fun):
            return Stream(map(lambda tup: fun(*tup), self.iterable))
        else:
            return Stream(map(fun, self.iterable))

    def filter(self, fun: Callable[[T], S]) -> "Stream[T]":
        """Filters elements using the supplied function. When iterating over tuples, the function can take multiple arguments."""
        if self.__should_expand(fun):
            return Stream(filter(lambda tup: fun(*tup), self.iterable))
        else:
            return Stream(filter(fun, self.iterable))

    def forEach(self, fun: Callable[[T], Any]) -> None:
        """Calls the function with each element. This is a terminal operation."""
        if self.__should_expand(fun):
            for i in self:
                fun(*i)
        else:
            for i in self:
                fun(i)

    def anyMatch(self, fun: Callable[[T], bool]) -> bool:
        """Returns True if any element of the stream matches the criteria."""
        return any(self.map(fun))

    def allMatch(self, fun: Callable[[T], bool]) -> bool:
        """Returns True if all elements of the stream match the criteria."""
        return all(self.map(fun))

    def noneMatch(self, fun: Callable[[T], bool]) -> bool:
        """Returns True if no element of the stream matches the criteria."""
        return not self.anyMatch(fun)

    def firstMatch(self, fun: Callable[[T], bool]) -> Nullable[T]:
        """Returns a Nullable of the first element matching the criteria. If none exist, returns an empty Nullable."""
        if self.__should_expand(fun):
            for i in self:
                if fun(*i):
                    return Nullable(i)
        else:
            for i in self:
                if fun(i):
                    return Nullable(i)
        return Nullable.empty()

    def flat(self: "Stream[Iterable[T]]") -> "Stream[T]":
        """When iterating over lists, flattens the stream by concatenating all lists."""
        return Stream(itertools.chain(*self))

    def toList(self) -> List[T]:
        """Collects all elements to a list."""
        return list(self)

    def toSet(self) -> Set[T]:
        """Collects all elements to a set."""
        return set(self)

    def toDict(self) -> dict:
        """When iterating over tuples, collects all elements to a dictionary. The first element becomes the key, the second the value."""
        return dict(self)

    def toTuple(self) -> Tuple[T]:
        """Collects all elements to a tuple."""
        return tuple(self)

    def __should_expand(self, fun: Callable) -> bool:
        if inspect.isbuiltin(fun):
            return False
        if inspect.isclass(fun):
            return len(inspect.signature(fun.__init__).parameters) > 2
        else:
            sig = inspect.signature(fun)
        return len(sig.parameters) > 1

    def count(self) -> int:
        """Returns the number of elements in the Stream. **Should never be used with an infinite stream!**"""
        if hasattr(self.iterable, '__len__'):
            return len(self.iterable)
        return self.reduce(0, lambda accumulator, element: accumulator + 1)

    def reduce(self, start_value: Any, reducer: Callable[[Any, T], Any]):
        """ Reduce using the supplied function.

        Args:
            start_value: starting value for the accumulator.
            reducer (Callable) : e.g. lambda accumulator, element: accumulator + element"""
        accumulator = start_value
        for element in self:
            accumulator = reducer(accumulator, element)
        return accumulator

    def __iter__(self) -> Iterator:
        return iter(self.iterable)

    @staticmethod
    def zip(*iterables: Iterable[T]) -> "Stream[Tuple[T]]":
        """Creates a stream by *zipping* the iterables, instead of concatenating them.

        Returns:
            Stream of tuples.
        """
        return Stream(zip(*iterables))

    def unzip(self) -> Tuple[tuple, ...]:
        """When iterating over tuples, unwraps the stream back to separate lists."""
        return tuple(zip(*self))

    @staticmethod
    def of(*args) -> "Stream":
        """Creates a stream with non iterable arguments.

        Examples:
            >>> Stream.of(1,2,3,4).toList()
            [1,2,3,4]
            """
        return Stream(args)

    def sum(self):
        """Returns the sum of all elements in the stream."""
        return sum(self)

    def min(self) -> T:
        """Returns the min of all elements in the stream."""
        return min(self)

    def max(self) -> T:
        """Returns the max of all elements in the stream."""
        return max(self)

    def take(self, number: int) -> "Stream[T]":
        """Limit the stream to a specific number of items.

        Examples:
            >>> Stream.range().take(5).toList()
            [0,1,2,3,4]
            """
        return Stream(itertools.islice(self, number))

    @staticmethod
    def range(*args) -> "Stream[int]":
        """Creates an incrementing, integer stream. If arguments are supplied, they are passed as-is to the builtin `range` function. Otherwise, an infinite stream is created, starting at 0."""
        if len(args) == 0:
            return Stream(itertools.count())
        else:
            return Stream(range(*args))

    def first(self) -> Nullable[T]:
        """Returns a nullable containing the first element of the stream. If the stream is empty, returns an empty nullable."""
        return Nullable(next(self.iterable, None))
